from abc import abstractmethod
from typing import Optional
from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ...ndimage import fftn, ifftn, rfftn
from .._config import AbstractConfig
from .._potential_integrator import AbstractPotentialIntegrator
from .._potential_representation import AbstractPotentialRepresentation
from .._transfer_theory import (
    ContrastTransferTheory,
    WaveTransferTheory,
)


class AbstractScatteringTheory(eqx.Module, strict=True):
    """Base class for a scattering theory."""

    integrator: eqx.AbstractVar[AbstractPotentialIntegrator]

    @abstractmethod
    def compute_contrast_spectrum(
        self,
        potential: AbstractPotentialRepresentation,
        config: AbstractConfig,
        rng_key: Optional[PRNGKeyArray] = None,
        defocus_offset: Optional[float | Float[Array, ""]] = None,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        raise NotImplementedError

    @abstractmethod
    def compute_intensity_spectrum(
        self,
        potential: AbstractPotentialRepresentation,
        config: AbstractConfig,
        rng_key: Optional[PRNGKeyArray] = None,
        defocus_offset: Optional[float | Float[Array, ""]] = None,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        raise NotImplementedError


class AbstractWaveScatteringTheory(AbstractScatteringTheory, strict=True):
    """Base class for a wave-based scattering theory."""

    transfer_theory: eqx.AbstractVar[WaveTransferTheory]
    amplitude_contrast_ratio: eqx.AbstractVar[Float[Array, ""]]

    @abstractmethod
    def compute_exit_wave(
        self,
        potential: AbstractPotentialRepresentation,
        config: AbstractConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim}"]:
        raise NotImplementedError

    @override
    def compute_intensity_spectrum(
        self,
        potential: AbstractPotentialRepresentation,
        config: AbstractConfig,
        rng_key: Optional[PRNGKeyArray] = None,
        defocus_offset: Optional[float | Float[Array, ""]] = None,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        # ... compute the exit wave
        fourier_wavefunction = fftn(self.compute_exit_wave(potential, config, rng_key))
        # ... propagate to the detector plane
        fourier_wavefunction = self.transfer_theory.propagate_exit_wave(
            fourier_wavefunction,
            config,
            defocus_offset=defocus_offset,
        )
        wavefunction = ifftn(fourier_wavefunction)
        # ... get the squared wavefunction and return to fourier space
        intensity_spectrum = rfftn((wavefunction * jnp.conj(wavefunction)).real)

        return intensity_spectrum

    @override
    def compute_contrast_spectrum(
        self,
        potential: AbstractPotentialRepresentation,
        config: AbstractConfig,
        rng_key: Optional[PRNGKeyArray] = None,
        defocus_offset: Optional[float | Float[Array, ""]] = None,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Compute the contrast at the detector plane, given the squared wavefunction."""
        # ... compute the exit wave
        fourier_wavefunction = fftn(self.compute_exit_wave(potential, config, rng_key))
        # ... propagate to the detector plane
        fourier_wavefunction = self.transfer_theory.propagate_exit_wave(
            fourier_wavefunction,
            config,
            defocus_offset=defocus_offset,
        )
        wavefunction = ifftn(fourier_wavefunction)
        # ... get the squared wavefunction
        squared_wavefunction = (wavefunction * jnp.conj(wavefunction)).real
        # ... compute the contrast directly from the squared wavefunction
        # as C = -1 + psi^2 / 1 + psi^2
        contrast_spectrum = rfftn(
            (-1 + squared_wavefunction) / (1 + squared_wavefunction)
        )

        return contrast_spectrum


class AbstractWeakPhaseScatteringTheory(AbstractScatteringTheory, strict=True):
    """Base class for a scattering theory in linear image formation theory
    (the weak-phase approximation).
    """

    transfer_theory: eqx.AbstractVar[ContrastTransferTheory]

    @abstractmethod
    def compute_object_spectrum(
        self,
        potential: AbstractPotentialRepresentation,
        config: AbstractConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        raise NotImplementedError

    @override
    def compute_intensity_spectrum(
        self,
        potential: AbstractPotentialRepresentation,
        config: AbstractConfig,
        rng_key: Optional[PRNGKeyArray] = None,
        defocus_offset: Optional[float | Float[Array, ""]] = None,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Compute the squared wavefunction at the detector plane, given the
        contrast.
        """
        N1, N2 = config.padded_shape
        # ... compute the squared wavefunction directly from the image contrast
        # as |psi|^2 = 1 + 2C.
        contrast_spectrum = self.compute_contrast_spectrum(
            potential, config, rng_key, defocus_offset=defocus_offset
        )
        intensity_spectrum = (2 * contrast_spectrum).at[0, 0].add(1.0 * N1 * N2)
        return intensity_spectrum



from typing import Optional
from typing_extensions import override

import jax.numpy as jnp
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ...internal import error_if_not_fractional
from ...ndimage import ifftn, irfftn
from .._common_functions import apply_amplitude_contrast_ratio, apply_interaction_constant
from .._config import AbstractConfig
from .._potential_integrator import AbstractDirectIntegrator
from .._potential_representation import AbstractPotentialRepresentation
from .._solvent import AbstractRandomSolvent
from .._transfer_theory import WaveTransferTheory
from .base_scattering_theory import AbstractWaveScatteringTheory


class HighEnergyScatteringTheory(AbstractWaveScatteringTheory, strict=True):
    """Scattering theory in the high-energy approximation (eikonal approximation).

    This is the simplest model for multiple scattering events.

    **References:**

    - For the definition of the exit wave in the eikonal approximation, see Chapter 69,
      Page 2012, from *Hawkes, Peter W., and Erwin Kasper. Principles of Electron
      Optics, Volume 4: Advanced Wave Optics. Academic Press, 2022.*
    """

    integrator: AbstractDirectIntegrator
    transfer_theory: WaveTransferTheory
    solvent: Optional[AbstractRandomSolvent]
    amplitude_contrast_ratio: Float[Array, ""]

    def __init__(
        self,
        integrator: AbstractDirectIntegrator,
        transfer_theory: WaveTransferTheory,
        solvent: Optional[AbstractRandomSolvent] = None,
        amplitude_contrast_ratio: float | Float[Array, ""] = 0.1,
    ):
        """**Arguments:**

        - `integrator`: The method for integrating the scattering potential.
        - `transfer_theory`: The wave transfer theory.
        - `solvent`: The model for the solvent.
        - `amplitude_contrast_ratio`: The amplitude contrast ratio.
        """
        self.integrator = integrator
        self.transfer_theory = transfer_theory
        self.solvent = solvent
        self.amplitude_contrast_ratio = error_if_not_fractional(amplitude_contrast_ratio)

    @override
    def compute_exit_wave(
        self,
        potential: AbstractPotentialRepresentation,
        config: AbstractConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim}"]:
        # Compute the integrated potential in the exit plane
        fourier_in_plane_potential = self.integrator.integrate(
            potential, config, outputs_real_space=False
        )
        # The integrated potential may not be from an rfft; this depends on
        # if it is a projection approx
        is_projection_approx = self.integrator.is_projection_approximation
        if rng_key is not None:
            # Get the potential of the specimen plus the ice
            if self.solvent is not None:
                fourier_in_plane_potential = self.solvent.compute_in_plane_potential(
                    rng_key,
                    fourier_in_plane_potential,
                    config,
                    input_is_rfft=is_projection_approx,
                )
        # Back to real-space; need to be careful if the object spectrum is not an
        # rfftn
        do_ifft = lambda ft: (
            irfftn(ft, s=config.padded_shape)
            if is_projection_approx
            else ifftn(ft, s=config.padded_shape)
        )
        integrated_potential = apply_amplitude_contrast_ratio(
            do_ifft(fourier_in_plane_potential), self.amplitude_contrast_ratio
        )
        object = apply_interaction_constant(
            integrated_potential, config.wavelength_in_angstroms
        )
        # Compute wavefunction, with amplitude and phase contrast
        return jnp.exp(1.0j * object)



"""
Voxel-based representations of the scattering potential.
"""

from abc import abstractmethod
from functools import cached_property
from typing import ClassVar, Optional, cast
from typing_extensions import Self, override

import equinox as eqx
import jax.numpy as jnp
import numpy as np
from equinox import AbstractClassVar, AbstractVar, field
from jaxtyping import Array, Complex, Float

from ...coordinates import make_coordinate_grid, make_frequency_slice
from ...internal import error_if_not_positive
from ...ndimage import (
    compute_spline_coefficients,
    crop_to_shape,
    fftn,
    pad_to_shape,
)
from ...ndimage.transforms import AbstractFilter
from .._pose import AbstractPose
from .base_potential import AbstractPotentialRepresentation


class AbstractVoxelPotential(AbstractPotentialRepresentation, strict=True):
    """Abstract interface for a voxel-based scattering potential representation."""

    voxel_size: AbstractVar[Float[Array, ""]]
    is_real_space: AbstractClassVar[bool]

    @property
    @abstractmethod
    def shape(self) -> tuple[int, ...]:
        """The shape of the voxel array."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
    ) -> Self:
        """Load an `AbstractVoxelPotential` from real-valued 3D electron
        scattering potential.
        """
        raise NotImplementedError


# Not public API
class AbstractFourierVoxelGridPotential(AbstractVoxelPotential, strict=True):
    """Abstract interface of a 3D scattering potential voxel grid
    in fourier-space.
    """

    frequency_slice_in_pixels: AbstractVar[Float[Array, "1 dim dim 3"]]

    @abstractmethod
    def __init__(
        self,
        fourier_voxel_grid: Complex[Array, "dim dim dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""] | float,
    ):
        raise NotImplementedError


class FourierVoxelGridPotential(AbstractFourierVoxelGridPotential):
    """A 3D scattering potential voxel grid in fourier-space."""

    fourier_voxel_grid: Complex[Array, "dim dim dim"]
    frequency_slice_in_pixels: Float[Array, "1 dim dim 3"]
    voxel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    is_real_space: ClassVar[bool] = False

    @override
    def __init__(
        self,
        fourier_voxel_grid: Complex[Array, "dim dim dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""] | float,
    ):
        """**Arguments:**

        - `fourier_voxel_grid`: The cubic voxel grid in fourier space.
        - `frequency_slice_in_pixels`: Frequency slice coordinate system.
        - `voxel_size`: The voxel size.
        """
        self.fourier_voxel_grid = jnp.asarray(fourier_voxel_grid)
        self.frequency_slice_in_pixels = frequency_slice_in_pixels
        self.voxel_size = jnp.asarray(voxel_size)

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the `fourier_voxel_grid`."""
        return cast(tuple[int, int, int], self.fourier_voxel_grid.shape)

    @cached_property
    def frequency_slice_in_angstroms(self) -> Float[Array, "1 dim dim 3"]:
        """The `frequency_slice_in_pixels` in angstroms."""
        return _safe_multiply_grid_by_constant(
            self.frequency_slice_in_pixels, 1 / self.voxel_size
        )

    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Return a new potential with a rotated `frequency_slice_in_pixels`."""
        return eqx.tree_at(
            lambda d: d.frequency_slice_in_pixels,
            self,
            pose.rotate_coordinates(self.frequency_slice_in_pixels, inverse=inverse),
        )

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        *,
        pad_scale: float = 1.0,
        pad_mode: str = "constant",
        filter: Optional[AbstractFilter] = None,
    ) -> Self:
        """Load an `AbstractFourierVoxelGridPotential` from real-valued 3D electron
        scattering potential voxel grid.

        **Arguments:**

        - `real_voxel_grid`: A scattering potential voxel grid in real space.
        - `voxel_size`: The voxel size of `real_voxel_grid`.
        - `pad_scale`: Scale factor at which to pad `real_voxel_grid` before fourier
                     transform. Must be a value greater than `1.0`.
        - `pad_mode`: Padding method. See `jax.numpy.pad` for documentation.
        - `filter`: A filter to apply to the result of the fourier transform of
                  `real_voxel_grid`, i.e. `fftn(real_voxel_grid)`. Note that the zero
                  frequency component is assumed to be in the corner.
        """
        # Cast to jax array
        real_voxel_grid, voxel_size = (
            jnp.asarray(real_voxel_grid),
            jnp.asarray(voxel_size),
        )
        # Pad template
        if pad_scale < 1.0:
            raise ValueError("`pad_scale` must be greater than 1.0")
        # ... always pad to even size to avoid interpolation issues in
        # fourier slice extraction.
        padded_shape = cast(
            tuple[int, int, int],
            tuple([int(s * pad_scale) for s in real_voxel_grid.shape]),
        )
        padded_real_voxel_grid = pad_to_shape(
            real_voxel_grid, padded_shape, mode=pad_mode
        )
        # Load potential and coordinates. For now, do not store the
        # fourier potential only on the half space. Fourier slice extraction
        # does not currently work if rfftn is used.
        fourier_voxel_grid_with_zero_in_corner = (
            fftn(padded_real_voxel_grid)
            if filter is None
            else filter(fftn(padded_real_voxel_grid))
        )
        # ... store the potential grid with the zero frequency component in the center
        fourier_voxel_grid = jnp.fft.fftshift(fourier_voxel_grid_with_zero_in_corner)
        # ... create in-plane frequency slice on the half space
        frequency_slice = make_frequency_slice(
            cast(tuple[int, int], padded_real_voxel_grid.shape[:-1]),
            outputs_rfftfreqs=False,
        )

        return cls(fourier_voxel_grid, frequency_slice, voxel_size)


class FourierVoxelSplinePotential(AbstractFourierVoxelGridPotential):
    """A 3D scattering potential voxel grid in fourier-space, represented
    by spline coefficients.
    """

    spline_coefficients: Complex[Array, "coeff_dim coeff_dim coeff_dim"]
    frequency_slice_in_pixels: Float[Array, "1 dim dim 3"]
    voxel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    is_real_space: ClassVar[bool] = False

    def __init__(
        self,
        spline_coefficients: Complex[Array, "coeff_dim coeff_dim coeff_dim"],
        frequency_slice_in_pixels: Float[Array, "1 dim dim 3"],
        voxel_size: Float[Array, ""] | float,
    ):
        """**Arguments:**

        - `spline_coefficients`:
            The spline coefficents computed from the cubic voxel grid
            in fourier space. See `cryojax.image.compute_spline_coefficients`.
        - `frequency_slice_in_pixels`:
            Frequency slice coordinate system.
            See `cryojax.coordinates.make_frequency_slice`.
        - `voxel_size`: The voxel size.
        """
        self.spline_coefficients = spline_coefficients
        self.frequency_slice_in_pixels = frequency_slice_in_pixels
        self.voxel_size = jnp.asarray(voxel_size)

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the original `fourier_voxel_grid` from which
        `coefficients` were computed.
        """
        return cast(
            tuple[int, int, int], tuple([s - 2 for s in self.spline_coefficients.shape])
        )

    @cached_property
    def frequency_slice_in_angstroms(self) -> Float[Array, "1 dim dim 3"]:
        """The `frequency_slice_in_pixels` in angstroms."""
        return _safe_multiply_grid_by_constant(
            self.frequency_slice_in_pixels, 1 / self.voxel_size
        )

    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Return a new potential with a rotated `frequency_slice_in_pixels`."""
        return eqx.tree_at(
            lambda d: d.frequency_slice_in_pixels,
            self,
            pose.rotate_coordinates(self.frequency_slice_in_pixels, inverse=inverse),
        )

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        *,
        pad_scale: float = 1.0,
        pad_mode: str = "constant",
        filter: Optional[AbstractFilter] = None,
    ) -> Self:
        """Load an `AbstractFourierVoxelGridPotential` from real-valued 3D electron
        scattering potential voxel grid.

        **Arguments:**

        - `real_voxel_grid`: A scattering potential voxel grid in real space.
        - `voxel_size`: The voxel size of `real_voxel_grid`.
        - `pad_scale`: Scale factor at which to pad `real_voxel_grid` before fourier
                     transform. Must be a value greater than `1.0`.
        - `pad_mode`: Padding method. See `jax.numpy.pad` for documentation.
        - `filter`: A filter to apply to the result of the fourier transform of
                  `real_voxel_grid`, i.e. `fftn(real_voxel_grid)`. Note that the zero
                  frequency component is assumed to be in the corner.
        """
        # Cast to jax array
        real_voxel_grid, voxel_size = (
            jnp.asarray(real_voxel_grid),
            jnp.asarray(voxel_size),
        )
        # Pad template
        if pad_scale < 1.0:
            raise ValueError("`pad_scale` must be greater than 1.0")
        # ... always pad to even size to avoid interpolation issues in
        # fourier slice extraction.
        padded_shape = cast(
            tuple[int, int, int],
            tuple([int(s * pad_scale) for s in real_voxel_grid.shape]),
        )
        padded_real_voxel_grid = pad_to_shape(
            real_voxel_grid, padded_shape, mode=pad_mode
        )
        # Load potential and coordinates. For now, do not store the
        # fourier potential only on the half space. Fourier slice extraction
        # does not currently work if rfftn is used.
        fourier_voxel_grid_with_zero_in_corner = (
            fftn(padded_real_voxel_grid)
            if filter is None
            else filter(fftn(padded_real_voxel_grid))
        )
        # ... store the potential grid with the zero frequency component in the center
        fourier_voxel_grid = jnp.fft.fftshift(fourier_voxel_grid_with_zero_in_corner)
        # ... compute spline coefficients
        spline_coefficients = compute_spline_coefficients(jnp.asarray(fourier_voxel_grid))
        # ... create in-plane frequency slice on the half space
        frequency_slice = make_frequency_slice(
            cast(tuple[int, int], padded_real_voxel_grid.shape[:-1]),
            outputs_rfftfreqs=False,
        )

        return cls(spline_coefficients, frequency_slice, voxel_size)


class RealVoxelGridPotential(AbstractVoxelPotential, strict=True):
    """Abstraction of a 3D scattering potential voxel grid in real-space."""

    real_voxel_grid: Float[Array, "dim dim dim"]
    coordinate_grid_in_pixels: Float[Array, "dim dim dim 3"]
    voxel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    is_real_space: ClassVar[bool] = True

    def __init__(
        self,
        real_voxel_grid: Float[Array, "dim dim dim"],
        coordinate_grid_in_pixels: Float[Array, "dim dim dim 3"],
        voxel_size: Float[Array, ""] | float,
    ):
        """**Arguments:**

        - `real_voxel_grid`: The voxel grid in fourier space.
        - `coordinate_grid_in_pixels`: A coordinate grid.
        - `voxel_size`: The voxel size.
        """
        self.real_voxel_grid = jnp.asarray(real_voxel_grid)
        self.coordinate_grid_in_pixels = coordinate_grid_in_pixels
        self.voxel_size = jnp.asarray(voxel_size)

    @property
    def shape(self) -> tuple[int, int, int]:
        """The shape of the `real_voxel_grid`."""
        return cast(tuple[int, int, int], self.real_voxel_grid.shape)

    @cached_property
    def coordinate_grid_in_angstroms(self) -> Float[Array, "dim dim dim 3"]:
        """The `coordinate_grid_in_pixels` in angstroms."""
        return _safe_multiply_grid_by_constant(
            self.coordinate_grid_in_pixels, self.voxel_size
        )

    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Return a new potential with a rotated
        `coordinate_grid_in_pixels`.
        """
        return eqx.tree_at(
            lambda d: d.coordinate_grid_in_pixels,
            self,
            pose.rotate_coordinates(self.coordinate_grid_in_pixels, inverse=inverse),
        )

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        *,
        coordinate_grid_in_pixels: Optional[Float[Array, "dim dim dim 3"]] = None,
        crop_scale: Optional[float] = None,
    ) -> Self:
        """Load a `RealVoxelGridPotential` from a real-valued 3D electron
        scattering potential voxel grid.

        **Arguments:**

        - `real_voxel_grid`: An electron scattering potential voxel grid in real space.
        - `voxel_size`: The voxel size of `real_voxel_grid`.
        - `crop_scale`: Scale factor at which to crop `real_voxel_grid`.
                        Must be a value greater than `1`.
        """
        # Cast to jax array
        real_voxel_grid, voxel_size = (
            jnp.asarray(real_voxel_grid),
            jnp.asarray(voxel_size),
        )
        # Make coordinates if not given
        if coordinate_grid_in_pixels is None:
            # Option for cropping template
            if crop_scale is not None:
                if crop_scale < 1.0:
                    raise ValueError("`crop_scale` must be greater than 1.0")
                cropped_shape = cast(
                    tuple[int, int, int],
                    tuple([int(s / crop_scale) for s in real_voxel_grid.shape[-3:]]),
                )
                real_voxel_grid = crop_to_shape(real_voxel_grid, cropped_shape)
            coordinate_grid_in_pixels = make_coordinate_grid(real_voxel_grid.shape[-3:])

        return cls(real_voxel_grid, coordinate_grid_in_pixels, voxel_size)


class RealVoxelCloudPotential(AbstractVoxelPotential, strict=True):
    """Abstraction of a 3D electron scattering potential voxel point cloud.

    !!! info
        This object is similar to the `RealVoxelGridPotential`. Instead
        of storing the whole voxel grid, a `RealVoxelCloudPotential` need
        only store points of non-zero scattering potential. Therefore,
        a `RealVoxelCloudPotential` stores a point cloud of scattering potential
        voxel values. Instantiating with the `from_real_voxel_grid` constructor
        will automatically mask points of zero scattering potential.
    """

    voxel_weights: Float[Array, " size"]
    coordinate_list_in_pixels: Float[Array, "size 3"]
    voxel_size: Float[Array, ""] = field(converter=error_if_not_positive)

    is_real_space: ClassVar[bool] = True

    def __init__(
        self,
        voxel_weights: Float[Array, " size"],
        coordinate_list_in_pixels: Float[Array, "size 3"],
        voxel_size: Float[Array, ""] | float,
    ):
        """**Arguments:**

        - `voxel_weights`: A point-cloud of voxel scattering potential values.
        - `coordinate_list_in_pixels`: Coordinate list for the `voxel_weights`.
        - `voxel_size`: The voxel size.
        """
        self.voxel_weights = jnp.asarray(voxel_weights)
        self.coordinate_list_in_pixels = coordinate_list_in_pixels
        self.voxel_size = jnp.asarray(voxel_size)

    @property
    def shape(self) -> tuple[int]:
        """The shape of `voxel_weights`."""
        return cast(tuple[int], self.voxel_weights.shape)

    @cached_property
    def coordinate_list_in_angstroms(self) -> Float[Array, "size 3"]:
        """The `coordinate_list_in_pixels` in angstroms."""
        return _safe_multiply_list_by_constant(
            self.coordinate_list_in_pixels, self.voxel_size
        )

    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Return a new potential with a rotated
        `coordinate_list_in_pixels`.
        """
        return eqx.tree_at(
            lambda d: d.coordinate_list_in_pixels,
            self,
            pose.rotate_coordinates(self.coordinate_list_in_pixels, inverse=inverse),
        )

    @classmethod
    def from_real_voxel_grid(
        cls,
        real_voxel_grid: Float[Array, "dim dim dim"] | Float[np.ndarray, "dim dim dim"],
        voxel_size: Float[Array, ""] | Float[np.ndarray, ""] | float,
        *,
        coordinate_grid_in_pixels: Optional[Float[Array, "dim dim dim 3"]] = None,
        rtol: float = 1e-05,
        atol: float = 1e-08,
        size: Optional[int] = None,
        fill_value: Optional[float] = None,
    ) -> Self:
        """Load an `RealVoxelCloudPotential` from a real-valued 3D electron
        scattering potential voxel grid.

        **Arguments:**

        - `real_voxel_grid`: An electron scattering potential voxel grid in real space.
        - `voxel_size`: The voxel size of `real_voxel_grid`.
        - `rtol`: Argument passed to `jnp.isclose`, used for masking
                  voxels of zero scattering potential.
        - `atol`: Argument passed to `jnp.isclose`, used for masking
                  voxels of zero scattering potential.
        - `size`: Argument passed to `jnp.where`, used for fixing the size
                  of the masked scattering potential. This argument is required
                  for using this function with a JAX transformation.
        - `fill_value`: Argument passed to `jnp.where`, used if `size` is specified and
                        the mask has fewer than the indicated number of elements.
        """
        # Cast to jax array
        real_voxel_grid, voxel_size = (
            jnp.asarray(real_voxel_grid),
            jnp.asarray(voxel_size),
        )
        # Make coordinates if not given
        if coordinate_grid_in_pixels is None:
            coordinate_grid_in_pixels = make_coordinate_grid(real_voxel_grid.shape)
        # ... mask zeros to store smaller arrays. This
        # option is not jittable.
        nonzero = jnp.where(
            ~jnp.isclose(real_voxel_grid, 0.0, rtol=rtol, atol=atol),
            size=size,
            fill_value=fill_value,
        )
        flat_potential = real_voxel_grid[nonzero]
        coordinate_list = coordinate_grid_in_pixels[nonzero]

        return cls(flat_potential, coordinate_list, voxel_size)


def _safe_multiply_grid_by_constant(
    grid: Float[Array, "z_dim y_dim x_dim 3"], constant: Float[Array, ""]
) -> Float[Array, "z_dim y_dim x_dim 3"]:
    """Multiplies a coordinate grid by a constant in a
    safe way for gradient computation.
    """
    return jnp.where(grid != 0.0, jnp.asarray(constant) * grid, 0.0)


def _safe_multiply_list_by_constant(
    coordinate_list: Float[Array, "size 3"], constant: Float[Array, ""]
) -> Float[Array, "size 3"]:
    """Multiplies a coordinate grid by a constant in a
    safe way for gradient computation.
    """
    return jnp.where(coordinate_list != 0.0, jnp.asarray(constant) * coordinate_list, 0.0)



"""
Methods for integrating the scattering potential directly onto the exit plane.
"""

from abc import abstractmethod
from typing import Generic, Optional, TypeVar

import jax.numpy as jnp
from equinox import AbstractClassVar, AbstractVar, error_if
from jaxtyping import Array, Complex, Float

from ....ndimage import maybe_rescale_pixel_size
from ..._config import AbstractConfig
from ..._potential_representation import AbstractVoxelPotential
from ..base_potential_integrator import AbstractPotentialIntegrator


PotentialT = TypeVar("PotentialT")
VoxelPotentialT = TypeVar("VoxelPotentialT", bound="AbstractVoxelPotential")


class AbstractDirectIntegrator(
    AbstractPotentialIntegrator, Generic[PotentialT], strict=True
):
    """Base class for a method of integrating a potential onto
    the exit plane.
    """

    is_projection_approximation: AbstractClassVar[bool]

    @abstractmethod
    def integrate(
        self,
        potential: PotentialT,
        config: AbstractConfig,
        outputs_real_space: bool = False,
    ) -> (
        Complex[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim//2+1}",
        ]
        | Complex[Array, "{config.padded_y_dim} {config.padded_x_dim}"]
        | Float[Array, "{config.padded_y_dim} {config.padded_x_dim}"]
    ):
        raise NotImplementedError


class AbstractDirectVoxelIntegrator(
    AbstractDirectIntegrator[VoxelPotentialT], strict=True
):
    """Base class for a method of integrating a voxel-based potential."""

    pixel_rescaling_mode: AbstractVar[Optional[str]]

    def _maybe_rescale_pixel_size(
        self,
        fourier_in_plane_potential: Array,
        potential: AbstractVoxelPotential,
        config: AbstractConfig,
        input_is_rfft: bool,
    ) -> Array:
        """Return the integrated potential in fourier space at the
        `config.pixel_size` and the `config.padded_shape.`
        """
        if self.pixel_rescaling_mode is None:
            fourier_in_plane_potential = error_if(
                fourier_in_plane_potential,
                ~jnp.isclose(potential.voxel_size, config.pixel_size),
                f"Tried to use {type(self).__name__} with `{type(potential).__name__}."
                f"voxel_size != {type(potential).__name__}.pixel_size`. If this is true, "
                f"`{type(self).__name__}.pixel_size_rescaling_method` must not be set to "
                f"`None`. Try setting `{type(self).__name__}.pixel_size_rescaling_method "
                "= 'bicubic'`.",
            )
            return fourier_in_plane_potential
        else:
            fourier_in_plane_potential = maybe_rescale_pixel_size(
                fourier_in_plane_potential,
                potential.voxel_size,
                config.pixel_size,
                input_is_real=False,
                input_is_rfft=input_is_rfft,
                shape_in_real_space=config.padded_shape,
            )
            return fourier_in_plane_potential



from .transfer_function import (
    AberratedAstigmaticCTF as AberratedAstigmaticCTF,
    AbstractCTF as AbstractCTF,
    NullCTF as NullCTF,
)
from .transfer_theory import (
    AbstractTransferTheory as AbstractTransferTheory,
    ContrastTransferTheory as ContrastTransferTheory,
    WaveTransferTheory as WaveTransferTheory,
)



import math
from typing import ClassVar, Optional
from typing_extensions import override

import jax
import jax.numpy as jnp
import jax.scipy as jsp
from jaxtyping import Array, Complex, Float

from ....constants import convert_variance_to_b_factor
from ....coordinates import make_1d_coordinate_grid
from ....ndimage import (
    downsample_to_shape_with_fourier_cropping,
    resize_with_crop_or_pad,
    rfftn,
)
from ..._config import AbstractConfig
from ..._potential_representation import (
    GaussianMixtureAtomicPotential,
    PengAtomicPotential,
)
from .base_direct_integrator import AbstractDirectIntegrator


class GaussianMixtureProjection(
    AbstractDirectIntegrator[GaussianMixtureAtomicPotential | PengAtomicPotential],
    strict=True,
):
    upsampling_factor: Optional[int]
    shape: Optional[tuple[int, int]]
    use_error_functions: bool
    n_batches: int

    is_projection_approximation: ClassVar[bool] = True
    requires_inverse_rotation: ClassVar[bool] = False

    def __init__(
        self,
        *,
        upsampling_factor: Optional[int] = None,
        shape: Optional[tuple[int, int]] = None,
        use_error_functions: bool = False,
        n_batches: int = 1,
    ):
        """**Arguments:**

        - `upsampling_factor`:
            The factor by which to upsample the computation of the images.
            If `upsampling_factor` is greater than 1, the images will be computed
            at a higher resolution and then downsampled to the original resolution.
            This can be useful for reducing aliasing artifacts in the images.
        - `shape`:
            The shape of the plane on which projections are computed before padding or
            cropping to the `AbstractConfig.padded_shape`. This argument is particularly
            useful if the `AbstractConfig.padded_shape` is much larger than the protein.
        - `use_error_functions`:
            If `True`, use error functions to evaluate the projected potential at
            a pixel to be the average value within the pixel using gaussian
            integrals. If `False`, the potential at a pixel will simply be evaluated
            as a gaussian.
        - `n_batches`:
            The number of batches over groups of atoms
            used to evaluate the projection.
            This is useful if GPU memory is exhausted. By default,
            `1`, which computes a projection for all atoms at once.
        """  # noqa: E501
        self.upsampling_factor = upsampling_factor
        self.shape = shape
        self.use_error_functions = use_error_functions
        self.n_batches = n_batches

    def __check_init__(self):
        if self.upsampling_factor is not None and self.upsampling_factor < 1:
            raise AttributeError(
                "`GaussianMixtureProjection.upsampling_factor` must "
                f"be greater than `1`. Got a value of {self.upsampling_factor}."
            )

    @override
    def integrate(
        self,
        potential: GaussianMixtureAtomicPotential | PengAtomicPotential,
        config: AbstractConfig,
        outputs_real_space: bool = False,
    ) -> (
        Complex[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim//2+1}",
        ]
        | Float[Array, "{config.padded_y_dim} {config.padded_x_dim}"]
    ):
        """Compute a projection from the atomic potential and transform it to Fourier
        space.

        **Arguments:**

        - `potential`: The atomic potential to project.
        - `config`: The configuration of the imaging instrument.

        **Returns:**

        The integrated potential in real or fourier space at the `AbstractConfig.padded_shape`.
        """  # noqa: E501
        # Grab the image configuration
        shape = config.padded_shape if self.shape is None else self.shape
        pixel_size = config.pixel_size
        if self.upsampling_factor is not None:
            u = self.upsampling_factor
            upsampled_pixel_size, upsampled_shape = (
                pixel_size / u,
                (
                    shape[0] * u,
                    shape[1] * u,
                ),
            )
        else:
            upsampled_pixel_size, upsampled_shape = pixel_size, shape
        # Grab the gaussian amplitudes and widths
        if isinstance(potential, PengAtomicPotential):
            gaussian_amplitudes = potential.scattering_factor_a
            gaussian_widths = potential.scattering_factor_b
            if potential.b_factors is not None:
                gaussian_widths += potential.b_factors[:, None]
        elif isinstance(potential, GaussianMixtureAtomicPotential):
            gaussian_amplitudes = potential.gaussian_amplitudes
            gaussian_widths = jnp.asarray(
                convert_variance_to_b_factor(potential.gaussian_variances)
            )
        else:
            raise ValueError(
                "Supported types for `potential` are `PengAtomicPotential` and "
                "`GaussianMixtureAtomicPotential`."
            )
        # Compute the projection
        in_plane_potential = _compute_projected_potential_from_atoms(
            upsampled_shape,
            upsampled_pixel_size,
            potential.atom_positions,
            gaussian_amplitudes,
            gaussian_widths,
            self.use_error_functions,
            self.n_batches,
        )
        if self.upsampling_factor is not None:
            # Downsample back to the original pixel size, rescaling so that the
            # downsampling produces an average in a given region, not a sum
            n_pixels, upsampled_n_pixels = math.prod(shape), math.prod(upsampled_shape)
            if self.shape is None:
                return downsample_to_shape_with_fourier_cropping(
                    in_plane_potential * (n_pixels / upsampled_n_pixels),
                    downsampled_shape=shape,
                    outputs_real_space=outputs_real_space,
                )
            else:
                in_plane_potential = downsample_to_shape_with_fourier_cropping(
                    in_plane_potential * (n_pixels / upsampled_n_pixels),
                    downsampled_shape=shape,
                    outputs_real_space=True,
                )
                in_plane_potential = resize_with_crop_or_pad(
                    in_plane_potential, config.padded_shape
                )
                return (
                    in_plane_potential
                    if outputs_real_space
                    else rfftn(in_plane_potential)
                )
        else:
            if self.shape is None:
                return (
                    in_plane_potential
                    if outputs_real_space
                    else rfftn(in_plane_potential)
                )
            else:
                in_plane_potential = resize_with_crop_or_pad(
                    in_plane_potential, config.padded_shape
                )
                return (
                    in_plane_potential
                    if outputs_real_space
                    else rfftn(in_plane_potential)
                )


def _compute_projected_potential_from_atoms(
    shape: tuple[int, int],
    pixel_size: Float[Array, ""],
    atom_positions: Float[Array, "n_atoms 3"],
    a: Float[Array, "n_atoms n_gaussians_per_atom"],
    b: Float[Array, "n_atoms n_gaussians_per_atom"],
    use_error_functions: bool,
    n_batches: int,
) -> Float[Array, "dim_y dim_x"]:
    # Make the grid on which to evaluate the result
    grid_x = make_1d_coordinate_grid(shape[1], pixel_size)
    grid_y = make_1d_coordinate_grid(shape[0], pixel_size)
    # Get function and pytree to compute potential over a batch of atoms
    xs = (atom_positions, a, b)
    compute_potential_for_atom_group = (
        lambda xs: _compute_projected_potential_from_atom_group(
            grid_x,
            grid_y,
            pixel_size,
            xs[0],
            xs[1],
            xs[2],
            use_error_functions,
        )
    )
    # Compute projection with a call to `jax.lax.map` in batches
    if n_batches > atom_positions.shape[0]:
        raise ValueError(
            "The `n_batches` when computing a projection must "
            "be an integer less than or equal to the number of atoms, "
            f"which is equal to {atom_positions.shape[0]}. Got "
            f"`n_batches = {n_batches}`."
        )
    elif n_batches == 1:
        projection = compute_potential_for_atom_group(xs)
    elif n_batches > 1:
        projection = jnp.sum(
            _batched_map_with_contraction(
                compute_potential_for_atom_group, xs, n_batches
            ),
            axis=0,
        )
    else:
        raise ValueError(
            "The `n_batches` when building a voxel grid must be an "
            "integer greater than or equal to 1."
        )
    return projection


def _compute_projected_potential_from_atom_group(
    grid_x: Float[Array, " dim_x"],
    grid_y: Float[Array, " dim_y"],
    pixel_size: Float[Array, ""],
    atom_positions: Float[Array, "n_atoms 3"],
    a: Float[Array, "n_atoms n_gaussians_per_atom"],
    b: Float[Array, "n_atoms n_gaussians_per_atom"],
    use_error_functions: bool,
) -> Float[Array, "dim_y dim_x"]:
    # Evaluate 1D gaussian integrals for each of x, y, and z dimensions

    if use_error_functions:
        result = _compute_gaussian_integrals_for_all_atoms(
            grid_x, grid_y, atom_positions, a, b, pixel_size
        )
    else:
        result = _compute_gaussians_for_all_atoms(grid_x, grid_y, atom_positions, a, b)
    gaussians_times_prefactor_x, gaussians_y = result
    projection = _compute_projected_potential_from_gaussians(
        gaussians_times_prefactor_x, gaussians_y
    )

    return projection


def _compute_projected_potential_from_gaussians(
    gaussians_per_interval_per_atom_x: Float[Array, "dim_x n_atoms n_gaussians_per_atom"],
    gaussians_per_interval_per_atom_y: Float[Array, "dim_y n_atoms n_gaussians_per_atom"],
) -> Float[Array, "dim_y dim_x"]:
    # Prepare matrices with dimensions of the number of atoms and the number of grid
    # points. There are as many matrices as number of gaussians per atom
    gauss_x = jnp.transpose(gaussians_per_interval_per_atom_x, (2, 1, 0))
    gauss_y = jnp.transpose(gaussians_per_interval_per_atom_y, (2, 0, 1))
    # Compute matrix multiplication then sum over the number of gaussians per atom
    return jnp.sum(jnp.matmul(gauss_y, gauss_x), axis=0)


def _compute_gaussian_integrals_for_all_atoms(
    grid_x: Float[Array, " dim_x"],
    grid_y: Float[Array, " dim_y"],
    atom_positions: Float[Array, "n_atoms 3"],
    a: Float[Array, "n_atoms n_gaussians_per_atom"],
    b: Float[Array, "n_atoms n_gaussians_per_atom"],
    pixel_size: Float[Array, ""],
) -> tuple[
    Float[Array, "dim_x n_atoms n_gaussians_per_atom"],
    Float[Array, "dim_y n_atoms n_gaussians_per_atom"],
]:
    """Evaluate 1D averaged gaussians in x, y, and z dimensions
    for each atom and each gaussian per atom.
    """
    # Define function to compute integrals for each dimension
    scaling = 2 * jnp.pi / jnp.sqrt(b)
    integration_kernel = lambda delta: (
        jsp.special.erf(scaling[None, :, :] * (delta + pixel_size)[:, :, None])
        - jsp.special.erf(scaling[None, :, :] * delta[:, :, None])
    )
    # Compute outer product of left edge of grid points minus atomic positions
    left_edge_grid_x, left_edge_grid_y = (
        grid_x - pixel_size / 2,
        grid_y - pixel_size / 2,
    )
    delta_x, delta_y = (
        left_edge_grid_x[:, None] - atom_positions[:, 0],
        left_edge_grid_y[:, None] - atom_positions[:, 1],
    )
    # Compute gaussian integrals for each grid point, each atom, and
    # each gaussian per atom
    gauss_x, gauss_y = (integration_kernel(delta_x), integration_kernel(delta_y))
    # Compute the prefactors for each atom and each gaussian per atom
    # for the potential
    prefactor = (4 * jnp.pi * a) / (2 * pixel_size) ** 2
    # Multiply the prefactor onto one of the gaussians for efficiency
    return prefactor * gauss_x, gauss_y


def _compute_gaussians_for_all_atoms(
    grid_x: Float[Array, " x_dim"],
    grid_y: Float[Array, " y_dim"],
    atom_positions: Float[Array, "n_atoms 3"],
    a: Float[Array, "n_atoms n_gaussians_per_atom"],
    b: Float[Array, "n_atoms n_gaussians_per_atom"],
) -> tuple[
    Float[Array, "dim_x n_atoms n_gaussians_per_atom"],
    Float[Array, "dim_y n_atoms n_gaussians_per_atom"],
]:
    b_inverse = 4.0 * jnp.pi / b
    gauss_x = jnp.exp(
        -jnp.pi
        * b_inverse[None, :, :]
        * ((grid_x[:, None] - atom_positions.T[0, :]) ** 2)[:, :, None]
    )
    gauss_y = jnp.exp(
        -jnp.pi
        * b_inverse[None, :, :]
        * ((grid_y[:, None] - atom_positions.T[1, :]) ** 2)[:, :, None]
    )
    prefactor = 4 * jnp.pi * a[None, :, :] * b_inverse[None, :, :]

    return prefactor * gauss_x, gauss_y


def _batched_map_with_contraction(fun, xs, n_batches):
    # ... reshape into an iterative dimension and a batching dimension
    batch_dim = jax.tree.leaves(xs)[0].shape[0]
    batch_size = batch_dim // n_batches
    xs_per_batch = jax.tree.map(
        lambda x: x[: batch_dim - batch_dim % batch_size, ...].reshape(
            (n_batches, batch_size, *x.shape[1:])
        ),
        xs,
    )
    # .. compute the result and reshape back into one leading dimension
    result = jax.lax.map(fun, xs_per_batch)
    # ... if the batch dimension is not divisible by the batch size, need
    # to take care of the remainder
    if batch_dim % batch_size != 0:
        remainder = fun(
            jax.tree.map(lambda x: x[batch_dim - batch_dim % batch_size :, ...], xs)
        )[None, ...]
        result = jnp.concatenate([result, remainder], axis=0)
    return result



from abc import abstractmethod
from typing import Generic, TypeVar

from jaxtyping import Array, Complex, Float

from ..._config import AbstractConfig
from ..base_potential_integrator import AbstractPotentialIntegrator


PotentialT = TypeVar("PotentialT")


class AbstractMultisliceIntegrator(
    AbstractPotentialIntegrator, Generic[PotentialT], strict=True
):
    """Base class for a multi-slice integration scheme."""

    @abstractmethod
    def integrate(
        self,
        potential: PotentialT,
        config: AbstractConfig,
        amplitude_contrast_ratio: Float[Array, ""] | float,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim}"]:
        raise NotImplementedError



from collections.abc import Callable
from typing import Optional

import jax
import tqdm.auto
import tqdm.notebook
import tqdm.std
from jax.debug import callback


def fori_loop_tqdm_decorator(
    n_iterations: int,
    print_every: Optional[int] = None,
    **kwargs,
) -> Callable:
    """Add a tqdm progress bar to `body_fun` used in `jax.lax.fori_loop`.
    This function is closely based on the implementation in
    [`jax_tqdm`](https://github.com/jeremiecoullon/jax-tqdm).
    """

    _update_progress_bar, close_tqdm = _build_tqdm(n_iterations, print_every, **kwargs)

    def _fori_loop_tqdm_decorator(func):
        def wrapper_progress_bar(i, val):
            _update_progress_bar(i)
            result = func(i, val)
            return close_tqdm(result, i)

        return wrapper_progress_bar

    return _fori_loop_tqdm_decorator


def _build_tqdm(
    n: int,
    print_rate: Optional[int] = None,
    tqdm_type: str = "auto",
    **kwargs,
) -> tuple[Callable, Callable]:
    if tqdm_type not in ("auto", "std", "notebook"):
        raise ValueError(
            'tqdm_type should be one of "auto", "std", or "notebook" '
            f'but got "{tqdm_type}"'
        )
    pbar = getattr(tqdm, tqdm_type).tqdm

    desc = kwargs.pop("desc", f"Running for {n:,} iterations")
    message = kwargs.pop("message", desc)
    for kwarg in ("total", "mininterval", "maxinterval", "miniters"):
        kwargs.pop(kwarg, None)

    tqdm_bars = {}

    if print_rate is None:
        if n > 20:
            print_rate = int(n / 20)
        else:
            print_rate = 1
    else:
        if print_rate < 1:
            raise ValueError(f"Print rate should be > 0 got {print_rate}")
        elif print_rate > n:
            raise ValueError(
                "Print rate should be less than the "
                f"number of steps {n}, got {print_rate}"
            )

    remainder = n % print_rate

    def _define_tqdm(arg, transform):
        tqdm_bars[0] = pbar(range(n), **kwargs)
        tqdm_bars[0].set_description(message, refresh=False)

    def _update_tqdm(arg, transform):
        tqdm_bars[0].update(int(arg))

    def _update_progress_bar(iter_num):
        "Updates tqdm from a JAX scan or loop"
        _ = jax.lax.cond(
            iter_num == 0,
            lambda _: callback(_define_tqdm, None, None, ordered=True),
            lambda _: None,
            operand=None,
        )

        _ = jax.lax.cond(
            # update tqdm every multiple of `print_rate` except at the end
            (iter_num % print_rate == 0) & (iter_num != n - remainder),
            lambda _: callback(_update_tqdm, print_rate, None, ordered=True),
            lambda _: None,
            operand=None,
        )

        _ = jax.lax.cond(
            # update tqdm by `remainder`
            iter_num == n - remainder,
            lambda _: callback(_update_tqdm, remainder, None, ordered=True),
            lambda _: None,
            operand=None,
        )

    def _close_tqdm(arg, transform):
        tqdm_bars[0].close()

    def close_tqdm(result, iter_num):
        _ = jax.lax.cond(
            iter_num == n - 1,
            lambda _: callback(_close_tqdm, None, None, ordered=True),
            lambda _: None,
            operand=None,
        )
        return result

    return _update_progress_bar, close_tqdm



"""
Base classes for image transformations.
"""

import equinox as eqx
import jax
from jaxtyping import Array, Inexact


class AbstractImageTransform(eqx.Module, strict=True):
    """Base class for computing and applying an ``Array`` to an image."""

    array: eqx.AbstractVar[
        Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]
    ]

    def get(self) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.array

    def __call__(
        self, image: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return image * jax.lax.stop_gradient(self.array)

    def __mul__(self, other) -> "AbstractImageTransform":
        return ProductImageTransform(transform1=self, transform2=other)

    def __rmul__(self, other) -> "AbstractImageTransform":
        return ProductImageTransform(transform1=other, transform2=self)


class ProductImageTransform(AbstractImageTransform, strict=True):
    """A helper to represent the product of two transforms."""

    array: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]

    transform1: AbstractImageTransform
    transform2: AbstractImageTransform

    def __init__(
        self,
        transform1: AbstractImageTransform,
        transform2: AbstractImageTransform,
    ):
        self.transform1 = transform1
        self.transform2 = transform2
        self.array = transform1.array * transform2.array

    def __repr__(self):
        return f"{repr(self.transform1)} * {repr(self.transform2)}"



from ._api_utils import make_image_model as make_image_model
from ._common_functions import (
    apply_amplitude_contrast_ratio as apply_amplitude_contrast_ratio,
    apply_interaction_constant as apply_interaction_constant,
)
from ._config import (
    AbstractConfig as AbstractConfig,
    BasicConfig as BasicConfig,
    DoseConfig as DoseConfig,
    GridHelper as GridHelper,
)
from ._detector import (
    AbstractDetector as AbstractDetector,
    AbstractDQE as AbstractDQE,
    CountingDQE as CountingDQE,
    GaussianDetector as GaussianDetector,
    NullDQE as NullDQE,
    PoissonDetector as PoissonDetector,
)
from ._distributions import (
    AbstractDistribution as AbstractDistribution,
    AbstractGaussianDistribution as AbstractGaussianDistribution,
    IndependentGaussianFourierModes as IndependentGaussianFourierModes,
    IndependentGaussianPixels as IndependentGaussianPixels,
)
from ._image_model import (
    AbstractImageModel as AbstractImageModel,
    AbstractPhysicalImageModel as AbstractPhysicalImageModel,
    ContrastImageModel as ContrastImageModel,
    ElectronCountsImageModel as ElectronCountsImageModel,
    IntensityImageModel as IntensityImageModel,
    LinearImageModel as LinearImageModel,
    ProjectionImageModel as ProjectionImageModel,
)
from ._pose import (
    AbstractPose as AbstractPose,
    AxisAnglePose as AxisAnglePose,
    EulerAnglePose as EulerAnglePose,
    QuaternionPose as QuaternionPose,
)
from ._potential_integrator import (
    AbstractDirectIntegrator as AbstractDirectIntegrator,
    AbstractDirectVoxelIntegrator as AbstractDirectVoxelIntegrator,
    AbstractPotentialIntegrator as AbstractPotentialIntegrator,
    FourierSliceExtraction as FourierSliceExtraction,
    GaussianMixtureProjection as GaussianMixtureProjection,
    NufftProjection as NufftProjection,
)
from ._potential_representation import (
    AbstractAtomicPotential as AbstractAtomicPotential,
    AbstractPotentialRepresentation as AbstractPotentialRepresentation,
    AbstractTabulatedAtomicPotential as AbstractTabulatedAtomicPotential,
    AbstractVoxelPotential as AbstractVoxelPotential,
    FourierVoxelGridPotential as FourierVoxelGridPotential,
    FourierVoxelSplinePotential as FourierVoxelSplinePotential,
    GaussianMixtureAtomicPotential as GaussianMixtureAtomicPotential,
    PengAtomicPotential as PengAtomicPotential,
    RealVoxelCloudPotential as RealVoxelCloudPotential,
    RealVoxelGridPotential as RealVoxelGridPotential,
)
from ._scattering_theory import (
    AbstractScatteringTheory as AbstractScatteringTheory,
    AbstractWeakPhaseScatteringTheory as AbstractWeakPhaseScatteringTheory,
    WeakPhaseScatteringTheory as WeakPhaseScatteringTheory,
)
from ._solvent import AbstractRandomSolvent as AbstractRandomSolvent
from ._structure import (
    AbstractStructuralEnsemble as AbstractStructuralEnsemble,
    AbstractStructure as AbstractStructure,
    BasicStructure as BasicStructure,
    DiscreteStructuralEnsemble as DiscreteStructuralEnsemble,
)
from ._transfer_theory import (
    AberratedAstigmaticCTF as AberratedAstigmaticCTF,
    AberratedAstigmaticCTF as CTF,  # noqa: F401
    AbstractCTF as AbstractCTF,
    AbstractTransferTheory as AbstractTransferTheory,
    ContrastTransferTheory as ContrastTransferTheory,
    NullCTF as NullCTF,
)



"""
Image formation models simulated from gaussian noise distributions.
"""

from abc import abstractmethod
from typing import Optional
from typing_extensions import override

import jax.numpy as jnp
import jax.random as jr
from equinox import AbstractVar
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ...internal import NDArrayLike, error_if_not_positive
from ...ndimage import rfftn
from ...ndimage.operators import Constant, FourierOperatorLike
from ...ndimage.transforms import FilterLike, MaskLike
from .._image_model import AbstractImageModel
from ._base_distribution import AbstractDistribution


RealImageArray = Float[
    Array,
    "{self.image_model.config.y_dim} {self.image_model.config.x_dim}",  # noqa: E501
]
FourierImageArray = Complex[
    Array,
    "{self.image_model.config.y_dim} {self.image_model.config.x_dim//2+1}",  # noqa: E501
]
ImageArray = RealImageArray | FourierImageArray


class AbstractGaussianDistribution(AbstractDistribution, strict=True):
    r"""An `AbstractDistribution` where images are formed via additive
    gaussian noise.

    Subclasses may compute the likelihood in real or fourier space and
    make different assumptions about the variance / covariance.
    """

    image_model: AbstractVar[AbstractImageModel]
    signal_scale_factor: AbstractVar[Float[Array, ""]]
    signal_offset: AbstractVar[Float[Array, ""]]

    @override
    def sample(
        self,
        rng_key: PRNGKeyArray,
        *,
        outputs_real_space: bool = True,
        mask: Optional[MaskLike] = None,
        filter: Optional[FilterLike] = None,
    ) -> ImageArray:
        """Sample a noisy image from the gaussian noise model.

        **Arguments:**

        - `outputs_real_space`:
            If `True`, return the signal in real space.
        - `mask`:
            A mask to apply to the final image and/or
            use for normalization.
        - `filter`:
            A filter to apply to the final image.
        """
        return self.compute_signal(
            outputs_real_space=outputs_real_space,
            mask=mask,
            filter=filter,
        ) + self.compute_noise(
            rng_key,
            outputs_real_space=outputs_real_space,
            mask=mask,
            filter=filter,
        )

    @override
    def compute_signal(
        self,
        *,
        outputs_real_space: bool = True,
        mask: Optional[MaskLike] = None,
        filter: Optional[FilterLike] = None,
    ) -> ImageArray:
        """Render the signal from the image formation model.

        **Arguments:**

        - `outputs_real_space`:
            If `True`, return the signal in real space.
        - `mask`:
            A mask to apply to the final image and/or
            use for normalization.
        - `filter`:
            A filter to apply to the final image.
        """
        simulated_image = self.image_model.simulate(
            outputs_real_space=True, mask=None, filter=filter
        )
        simulated_image = self.signal_scale_factor * simulated_image + self.signal_offset
        if mask is not None:
            simulated_image = mask(simulated_image)
        return simulated_image if outputs_real_space else rfftn(simulated_image)

    @abstractmethod
    def compute_noise(
        self,
        rng_key: PRNGKeyArray,
        *,
        outputs_real_space: bool = True,
        mask: Optional[MaskLike] = None,
        filter: Optional[FilterLike] = None,
    ) -> ImageArray:
        """Draw a realization from the gaussian noise model and return either in
        real or fourier space.
        """
        raise NotImplementedError


class IndependentGaussianPixels(AbstractGaussianDistribution, strict=True):
    r"""A gaussian noise model, where each pixel is independently drawn from
    a zero-mean gaussian of fixed variance (white noise).

    This computes the likelihood in real space, where the variance is a
    constant value across all pixels.
    """

    image_model: AbstractImageModel
    variance: Float[Array, ""]
    signal_scale_factor: Float[Array, ""]
    signal_offset: Float[Array, ""]

    def __init__(
        self,
        image_model: AbstractImageModel,
        variance: float | Float[NDArrayLike, ""] = 1.0,
        signal_scale_factor: float | Float[NDArrayLike, ""] = 1.0,
        signal_offset: float | Float[NDArrayLike, ""] = 0.0,
    ):
        """**Arguments:**

        - `image_model`:
            The image formation model.
        - `variance`:
            The variance of each pixel.
        - `signal_scale_factor`:
            A scale factor for the underlying signal simulated
            from `image_model`.
        - `signal_offset`:
            An offset for the underlying signal simulated from `image_model`.
        """  # noqa: E501
        self.image_model = image_model
        self.variance = error_if_not_positive(jnp.asarray(variance, dtype=float))
        self.signal_scale_factor = error_if_not_positive(
            jnp.asarray(signal_scale_factor, dtype=float)
        )
        self.signal_offset = jnp.asarray(signal_offset, dtype=float)

    @override
    def compute_noise(
        self,
        rng_key: PRNGKeyArray,
        *,
        outputs_real_space: bool = True,
        mask: Optional[MaskLike] = None,
        filter: Optional[FilterLike] = None,
    ) -> ImageArray:
        """Sample a realization of the noise from the distribution.

        **Arguments:**

        - `outputs_real_space`:
            If `True`, return the noise in real space.
        - `mask`:
            A mask to apply to the final image and/or
            use for normalization.
        - `filter`:
            A filter to apply to the final image.
        """
        pipeline = self.image_model
        n_pixels = pipeline.config.padded_n_pixels
        freqs = pipeline.config.padded_frequency_grid_in_angstroms
        # Compute the zero mean variance and scale up to be independent of the number of
        # pixels
        std = jnp.sqrt(n_pixels * self.variance)
        noise = pipeline.postprocess(
            std
            * jr.normal(rng_key, shape=freqs.shape[0:-1])
            .at[0, 0]
            .set(0.0)
            .astype(complex),
            outputs_real_space=outputs_real_space,
            mask=mask,
            filter=filter,
        )

        return noise

    @override
    def log_likelihood(
        self,
        observed: Float[
            Array,
            "{self.image_model.config.y_dim} " "{self.image_model.config.x_dim}",
        ],
        *,
        mask: Optional[MaskLike] = None,
        filter: Optional[FilterLike] = None,
    ) -> Float[Array, ""]:
        """Evaluate the log-likelihood of the gaussian noise model.

        !!! info

            When computing the likelihood, the observed image is assumed to have already
            been preprocessed with `filter` and `mask`. In other words,
            if `filter` or `mask` are `None`, these
            will *not* be applied to `observed`. The user must do this
            manually if desired.

        **Arguments:**

        - `observed` : The observed data in real space.
        - `mask`:
            A mask to apply to the final image and/or
            use for normalization.
        - `filter`:
            A filter to apply to the final image.
        """
        variance = self.variance
        # Create simulated data
        simulated = self.compute_signal(
            outputs_real_space=True,
            mask=mask,
            filter=filter,
        )
        # Compute residuals
        residuals = simulated - observed
        # Compute standard normal random variables
        squared_standard_normal_per_pixel = jnp.abs(residuals) ** 2 / (2 * variance)
        # Compute the log-likelihood for each pixel.
        log_likelihood_per_pixel = -1.0 * (
            squared_standard_normal_per_pixel + jnp.log(2 * jnp.pi * variance) / 2
        )
        # Compute log-likelihood, summing over pixels
        log_likelihood = jnp.sum(log_likelihood_per_pixel)

        return log_likelihood


class IndependentGaussianFourierModes(AbstractGaussianDistribution, strict=True):
    r"""A gaussian noise model, where each fourier mode is independent.

    This computes the likelihood in Fourier space,
    so that the variance to be an arbitrary noise power spectrum.
    """

    image_model: AbstractImageModel
    variance_function: FourierOperatorLike
    signal_scale_factor: Float[Array, ""]
    signal_offset: Float[Array, ""]

    def __init__(
        self,
        image_model: AbstractImageModel,
        variance_function: Optional[FourierOperatorLike] = None,
        signal_scale_factor: float | Float[NDArrayLike, ""] = 1.0,
        signal_offset: float | Float[NDArrayLike, ""] = 0.0,
    ):
        """**Arguments:**

        - `image_model`:
            The image formation model.
        - `variance_function`:
            The variance of each fourier mode. By default,
            `cryojax.image.operators.Constant(1.0)`.
        - `signal_scale_factor`:
            A scale factor for the underlying signal simulated from `image_model`.
        - `signal_offset`:
            An offset for the underlying signal simulated from `image_model`.
        """  # noqa: E501
        self.image_model = image_model
        self.variance_function = variance_function or Constant(1.0)
        self.signal_scale_factor = error_if_not_positive(
            jnp.asarray(signal_scale_factor, dtype=float)
        )
        self.signal_offset = jnp.asarray(signal_offset, dtype=float)

    def compute_noise(
        self,
        rng_key: PRNGKeyArray,
        *,
        outputs_real_space: bool = True,
        mask: Optional[MaskLike] = None,
        filter: Optional[FilterLike] = None,
    ) -> ImageArray:
        """Sample a realization of the noise from the distribution.

        **Arguments:**

        - `outputs_real_space`:
            If `True`, return the noise in real space.
        - `mask`:
            A mask to apply to the final image and/or
            use for normalization.
        - `filter`:
            A filter to apply to the final image.
        """
        pipeline = self.image_model
        n_pixels = pipeline.config.padded_n_pixels
        freqs = pipeline.config.padded_frequency_grid_in_angstroms
        # Compute the zero mean variance and scale up to be independent of the number of
        # pixels
        std = jnp.sqrt(n_pixels * self.variance_function(freqs))
        noise = pipeline.postprocess(
            std
            * jr.normal(rng_key, shape=freqs.shape[0:-1])
            .at[0, 0]
            .set(0.0)
            .astype(complex),
            outputs_real_space=outputs_real_space,
            mask=mask,
            filter=filter,
        )

        return noise

    @override
    def log_likelihood(
        self,
        observed: Complex[
            Array,
            "{self.image_model.config.y_dim} " "{self.image_model.config.x_dim//2+1}",
        ],
        *,
        mask: Optional[MaskLike] = None,
        filter: Optional[FilterLike] = None,
    ) -> Float[Array, ""]:
        """Evaluate the log-likelihood of the gaussian noise model.

        !!! info

            When computing the likelihood, the observed image is assumed to have already
            been preprocessed with `filter` and `mask`. In other words,
            if `filter` or `mask` are `None`, these
            will *not* be applied to `observed`. The user must do this
            manually if desired.

        **Arguments:**

        - `observed` : The observed data in fourier space.
        - `mask`:
            A mask to apply to the final image and/or
            use for normalization.
        - `filter`:
            A filter to apply to the final image.
        """
        pipeline = self.image_model
        n_pixels = pipeline.config.n_pixels
        freqs = pipeline.config.frequency_grid_in_angstroms
        # Compute the variance and scale up to be independent of the number of pixels
        variance = n_pixels * self.variance_function(freqs)
        # Create simulated data
        simulated = self.compute_signal(
            outputs_real_space=False,
            mask=mask,
            filter=filter,
        )
        # Compute residuals
        residuals = simulated - observed
        # Compute standard normal random variables
        squared_standard_normal_per_mode = jnp.abs(residuals) ** 2 / (2 * variance)
        # Compute the log-likelihood for each fourier mode.
        log_likelihood_per_mode = (
            squared_standard_normal_per_mode + jnp.log(2 * jnp.pi * variance) / 2
        )
        # Compute log-likelihood, throwing away the zero mode. Need to take care
        # to compute the loss function in fourier space for a real-valued function.
        log_likelihood = (
            -1.0
            * (
                jnp.sum(log_likelihood_per_mode[1:, 0])
                + 2 * jnp.sum(log_likelihood_per_mode[:, 1:])
            )
            / n_pixels
        )

        return log_likelihood



"""
Abstractions of ensembles on discrete conformational variables.
"""

from typing_extensions import override

import jax
import jax.numpy as jnp
from jaxtyping import Array, Int

from ...internal import error_if_negative
from .._pose import AbstractPose, EulerAnglePose
from .._potential_representation import AbstractPotentialRepresentation
from .base_ensemble import AbstractStructuralEnsemble


class DiscreteStructuralEnsemble(AbstractStructuralEnsemble, strict=True):
    """Abstraction of an ensemble with discrete conformational
    heterogeneity.
    """

    conformational_space: tuple[AbstractPotentialRepresentation, ...]
    pose: AbstractPose
    conformation: Int[Array, ""]

    def __init__(
        self,
        conformational_space: tuple[AbstractPotentialRepresentation, ...],
        pose: AbstractPose,
        conformation: int | Int[Array, ""],
    ):
        """**Arguments:**

        - `conformational_space`: A tuple of `AbstractPotential` representations.
        - `pose`: The pose of the specimen.
        - `conformation`: A conformation with a discrete index at which to evaluate
                          the scattering potential tuple.
        """
        self.conformational_space = conformational_space
        self.pose = pose or EulerAnglePose()
        self.conformation = jnp.asarray(error_if_negative(conformation))

    @override
    def get_potential_in_body_frame(self) -> AbstractPotentialRepresentation:
        """Get the scattering potential at configured conformation."""
        funcs = [
            lambda i=i: self.conformational_space[i]
            for i in range(len(self.conformational_space))
        ]
        potential = jax.lax.switch(self.conformation, funcs)

        return potential



from typing import Optional
from typing_extensions import override

from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ...internal import error_if_not_fractional
from .._config import AbstractConfig
from .._potential_integrator import AbstractMultisliceIntegrator
from .._potential_representation import AbstractPotentialRepresentation
from .._transfer_theory import WaveTransferTheory
from .base_scattering_theory import AbstractWaveScatteringTheory


class MultisliceScatteringTheory(AbstractWaveScatteringTheory, strict=True):
    """A scattering theory using the multislice method."""

    integrator: AbstractMultisliceIntegrator
    transfer_theory: WaveTransferTheory
    amplitude_contrast_ratio: Float[Array, ""]

    def __init__(
        self,
        integrator: AbstractMultisliceIntegrator,
        transfer_theory: WaveTransferTheory,
        amplitude_contrast_ratio: float | Float[Array, ""] = 0.1,
    ):
        """**Arguments:**

        - `integrator`: The multislice method.
        - `transfer_theory`: The wave transfer theory.
        - `amplitude_contrast_ratio`: The amplitude contrast ratio.
        """
        self.integrator = integrator
        self.transfer_theory = transfer_theory
        self.amplitude_contrast_ratio = error_if_not_fractional(amplitude_contrast_ratio)

    @override
    def compute_exit_wave(
        self,
        potential: AbstractPotentialRepresentation,
        config: AbstractConfig,
        rng_key: Optional[PRNGKeyArray] = None,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim}"]:
        # Compute the wavefunction in the exit plane
        wavefunction = self.integrator.integrate(
            potential, config, self.amplitude_contrast_ratio
        )

        return wavefunction



import os

import pytest


# jax.config.update("jax_numpy_dtype_promotion", "strict")
# jax.config.update("jax_numpy_rank_promotion", "raise")
# jax.config.update("jax_enable_x64", True)


@pytest.fixture
def sample_mrc_path():
    return os.path.join(os.path.dirname(__file__), "data", "3j9g_potential_ps4_4.mrc")


@pytest.fixture
def sample_pdb_path():
    return os.path.join(os.path.dirname(__file__), "data", "1uao.pdb")


@pytest.fixture
def sample_pdb_path_assembly():
    return os.path.join(os.path.dirname(__file__), "data", "1uao_assembly.pdb")


@pytest.fixture
def sample_cif_path():
    return os.path.join(os.path.dirname(__file__), "data", "8zpm.cif")


@pytest.fixture
def sample_pdb_url():
    return "https://files.rcsb.org/download/2XJX.pdb"


@pytest.fixture
def sample_waterbox_pdb():
    return os.path.join(os.path.dirname(__file__), "data", "relaxed_small_box_tip3p.pdb")



import jax
import numpy as np
import pytest

import cryojax.simulator as cxs
from cryojax.io import read_array_from_mrc
from cryojax.ndimage import crop_to_shape


jax.config.update("jax_enable_x64", True)


@pytest.fixture
def voxel_potential(sample_mrc_path):
    real_voxel_grid, voxel_size = read_array_from_mrc(sample_mrc_path, loads_spacing=True)
    return cxs.FourierVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size, pad_scale=1.3
    )


@pytest.fixture
def basic_config(voxel_potential):
    shape = voxel_potential.shape[0:2]
    return cxs.BasicConfig(
        shape=(int(0.9 * shape[0]), int(0.9 * shape[1])),
        pixel_size=voxel_potential.voxel_size,
        voltage_in_kilovolts=300.0,
        pad_options=dict(shape=shape),
    )


@pytest.fixture
def image_model(voxel_potential, basic_config):
    image_model = cxs.make_image_model(
        voxel_potential,
        basic_config,
        pose=cxs.EulerAnglePose(),
        transfer_theory=cxs.ContrastTransferTheory(cxs.AberratedAstigmaticCTF()),
    )
    return image_model


# Test correct image shape
@pytest.mark.parametrize("model", ["image_model"])
def test_real_shape(model, request):
    """Make sure shapes are as expected in real space."""
    model = request.getfixturevalue(model)
    image = model.simulate()
    padded_image = model.simulate(removes_padding=False)
    assert image.shape == model.config.shape
    assert padded_image.shape == model.config.padded_shape


@pytest.mark.parametrize("model", ["image_model"])
def test_fourier_shape(model, request):
    """Make sure shapes are as expected in fourier space."""
    model = request.getfixturevalue(model)
    image = model.simulate(outputs_real_space=False)
    padded_image = model.simulate(removes_padding=False, outputs_real_space=False)
    assert image.shape == model.config.frequency_grid_in_pixels.shape[0:2]
    assert padded_image.shape == model.config.padded_frequency_grid_in_pixels.shape[0:2]


@pytest.mark.parametrize("extra_dim_y, extra_dim_x", [(1, 1), (1, 0), (0, 1)])
def test_even_vs_odd_image_shape(extra_dim_y, extra_dim_x, voxel_potential):
    control_shape = voxel_potential.shape[0:2]
    test_shape = (control_shape[0] + extra_dim_y, control_shape[1] + extra_dim_x)
    config_control = cxs.BasicConfig(
        control_shape, pixel_size=voxel_potential.voxel_size, voltage_in_kilovolts=300.0
    )
    config_test = cxs.BasicConfig(
        test_shape, pixel_size=voxel_potential.voxel_size, voltage_in_kilovolts=300.0
    )
    pose = cxs.EulerAnglePose()
    transfer_theory = cxs.ContrastTransferTheory(cxs.AberratedAstigmaticCTF())
    model_control = cxs.make_image_model(
        voxel_potential, config_control, pose=pose, transfer_theory=transfer_theory
    )
    model_test = cxs.make_image_model(
        voxel_potential, config_test, pose=pose, transfer_theory=transfer_theory
    )
    np.testing.assert_allclose(
        crop_to_shape(model_test.simulate(), control_shape),
        model_control.simulate(),
        atol=1e-4,
    )



"""
Atomistic representation of the scattering potential.
"""

from abc import abstractmethod
from typing import Callable, Optional
from typing_extensions import Self, override

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.scipy as jsp
import jax.tree_util as jtu
import numpy as np
from jaxtyping import Array, Float, PyTree

from ...constants._conventions import convert_variance_to_b_factor
from ...coordinates import make_1d_coordinate_grid
from ...internal import error_if_negative, error_if_not_positive
from .._pose import AbstractPose
from .base_potential import AbstractPotentialRepresentation


class AbstractAtomicPotential(AbstractPotentialRepresentation, strict=True):
    """Abstract interface for an atom-based scattering potential representation.

    !!! info
        In, `cryojax`, potentials should be built in units of *inverse length squared*,
        $[L]^{-2}$. This rescaled potential is defined to be

        $$U(\\mathbf{r}) = \\frac{2 m e}{\\hbar^2} V(\\mathbf{r}),$$

        where $V$ is the electrostatic potential energy, $\\mathbf{r}$ is a positional
        coordinate, $m$ is the electron mass, and $e$ is the electron charge.

        For a single atom, this rescaled potential has the advantage that under usual
        scattering approximations (i.e. the first-born approximation), the
        fourier transform of this quantity is closely related to tabulated electron scattering
        factors. In particular, for a single atom with scattering factor $f^{(e)}(\\mathbf{q})$
        and scattering vector $\\mathbf{q}$, its rescaled potential is equal to

        $$U(\\mathbf{r}) = 4 \\pi \\mathcal{F}^{-1}[f^{(e)}(\\boldsymbol{\\xi} / 2)](\\mathbf{r}),$$

        where $\\boldsymbol{\\xi} = 2 \\mathbf{q}$ is the wave vector coordinate and
        $\\mathcal{F}^{-1}$ is the inverse fourier transform operator in the convention

        $$\\mathcal{F}[f](\\boldsymbol{\\xi}) = \\int d^3\\mathbf{r} \\ \\exp(2\\pi i \\boldsymbol{\\xi}\\cdot\\mathbf{r}) f(\\mathbf{r}).$$

        The rescaled potential $U$ gives the following time-independent schrodinger equation
        for the scattering problem,

        $$(\\nabla^2 + k^2) \\psi(\\mathbf{r}) = - U(\\mathbf{r}) \\psi(\\mathbf{r}),$$

        where $k$ is the incident wavenumber of the electron beam.

        **References**:

        - For the definition of the rescaled potential, see
        Chapter 69, Page 2003, Equation 69.6 from *Hawkes, Peter W., and Erwin Kasper.
        Principles of Electron Optics, Volume 4: Advanced Wave Optics. Academic Press,
        2022.*
        - To work out the correspondence between the rescaled potential and the electron
        scattering factors, see the supplementary information from *Vulović, Miloš, et al.
        "Image formation modeling in cryo-electron microscopy." Journal of structural
        biology 183.1 (2013): 19-32.*
    """  # noqa: E501

    atom_positions: eqx.AbstractVar[Float[Array, "n_atoms 3"]]

    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Return a new potential with rotated `atom_positions`."""
        return eqx.tree_at(
            lambda d: d.atom_positions,
            self,
            pose.rotate_coordinates(self.atom_positions, inverse=inverse),
        )

    def translate_to_pose(self, pose: AbstractPose) -> Self:
        """Return a new potential with rotated `atom_positions`."""
        offset_in_angstroms = pose.offset_in_angstroms
        if pose.offset_z_in_angstroms is None:
            offset_in_angstroms = jnp.concatenate(
                (offset_in_angstroms, jnp.atleast_1d(0.0))
            )
        return eqx.tree_at(
            lambda d: d.atom_positions, self, self.atom_positions + offset_in_angstroms
        )

    @abstractmethod
    def as_real_voxel_grid(
        self,
        shape: tuple[int, int, int],
        voxel_size: Float[Array, ""] | float,
    ) -> Float[Array, "{shape[0]} {shape[1]} {shape[2]}"]:
        raise NotImplementedError


class GaussianMixtureAtomicPotential(AbstractAtomicPotential, strict=True):
    r"""An atomistic representation of scattering potential as a mixture of
    gaussians.

    The naming and numerical convention of parameters `gaussian_amplitudes` and
    `gaussian_variances` follows "Robust Parameterization of Elastic and Absorptive
    Electron Atomic Scattering Factors" by Peng et al. (1996), where $a_i$ are
    the `gaussian_amplitudes` and $b_i$ are the `gaussian_variances` multiplied by $8\pi^2$.

    !!! info
        In order to load a `GaussianMixtureAtomicPotential` from tabulated
        scattering factors, use the `cryojax.constants` submodule.

        ```python
        from cryojax.constants import (
            read_peng_element_scattering_factor_parameter_table,
            get_tabulated_scattering_factor_parameters,
        )
        from cryojax.io import read_atoms_from_pdb
        from cryojax.simulator import GaussianMixtureAtomicPotential

        # Load positions of atoms and one-hot encoded atom names
        atom_positions, atom_identities = read_atoms_from_pdb(...)
        scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
            atom_identities, read_peng_element_scattering_factor_parameter_table()
        )
        potential = GaussianMixtureAtomicPotential(
            atom_positions,
            gaussian_amplitudes=scattering_factor_parameters["a"],
            gaussian_variances=scattering_factor_parameters["b"] / (8 * jnp.pi**2),
        )
        ```
    """  # noqa: E501

    atom_positions: Float[Array, "n_atoms 3"]
    gaussian_amplitudes: Float[Array, "n_atoms n_gaussians_per_atom"]
    gaussian_variances: Float[Array, "n_atoms n_gaussians_per_atom"]

    def __init__(
        self,
        atom_positions: Float[Array, "n_atoms 3"] | Float[np.ndarray, "n_atoms 3"],
        gaussian_amplitudes: (
            Float[Array, "n_atoms n_gaussians_per_atom"]
            | Float[np.ndarray, "n_atoms n_gaussians_per_atom"]
        ),
        gaussian_variances: (
            Float[Array, "n_atoms n_gaussians_per_atom"]
            | Float[np.ndarray, "n_atoms n_gaussians_per_atom"]
        ),
    ):
        """**Arguments:**

        - `atom_positions`: The coordinates of the atoms in units of angstroms.
        - `gaussian_amplitudes`:
            The strength for each atom and for each gaussian per atom.
            This has units of angstroms.
        - `gaussian_variances`:
            The variance (up to numerical constants) for each atom and
            for each gaussian per atom. This has units of angstroms
            squared.
        """
        self.atom_positions = jnp.asarray(atom_positions)
        self.gaussian_amplitudes = jnp.asarray(gaussian_amplitudes)
        self.gaussian_variances = error_if_not_positive(jnp.asarray(gaussian_variances))

    @override
    def as_real_voxel_grid(
        self,
        shape: tuple[int, int, int],
        voxel_size: Float[Array, ""] | float,
        *,
        batch_size_for_z_planes: int = 1,
        n_batches_of_atoms: int = 1,
    ) -> Float[Array, "{shape[0]} {shape[1]} {shape[2]}"]:
        """Return a voxel grid of the potential in real space.

        See [`PengAtomicPotential.as_real_voxel_grid`](potential.md#cryojax.simulator.PengAtomicPotential.as_real_voxel_grid)
        for the numerical conventions used when computing the sum of gaussians.

        **Arguments:**

        - `shape`: The shape of the resulting voxel grid.
        - `voxel_size`: The voxel size of the resulting voxel grid.
        - `batch_size_for_z_planes`:
            The number of z-planes to evaluate in parallel with
            `jax.vmap`. By default, `1`.
        - `n_batches_of_atoms`:
            The number of iterations used to evaluate the volume,
            where the iteration is taken over groups of atoms.
            This is useful if `batch_size_for_z_planes = 1`
            and GPU memory is exhausted. By default, `1`.

        **Returns:**

        The rescaled potential $U_{\\ell}$ as a voxel grid of shape `shape`
        and voxel size `voxel_size`.
        """  # noqa: E501
        return _build_real_space_voxel_potential_from_atoms(
            shape,
            jnp.asarray(voxel_size),
            self.atom_positions,
            self.gaussian_amplitudes,
            jnp.asarray(convert_variance_to_b_factor(self.gaussian_variances)),
            batch_size_for_z_planes=batch_size_for_z_planes,
            n_batches_of_atoms=n_batches_of_atoms,
        )


class AbstractTabulatedAtomicPotential(AbstractAtomicPotential, strict=True):
    b_factors: eqx.AbstractVar[Optional[Float[Array, " n_atoms"]]]


class PengAtomicPotential(AbstractTabulatedAtomicPotential, strict=True):
    """The scattering potential parameterized as a mixture of five gaussians
    per atom, through work by Lian-Mao Peng.

    To load this object, the following pattern can be used:

    ```python
    from cryojax.io import read_atoms_from_pdb
    from cryojax.simulator import PengAtomicPotential

    # Load positions of atoms and one-hot encoded atom names
    filename = "example.pdb"
    atom_positions, atom_identities = read_atoms_from_pdb(filename)
    scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
        atom_identities, read_peng_element_scattering_factor_parameter_table()
        )
    potential = PengAtomicPotential(
        atom_positions,
        scattering_factor_a=scattering_factor_parameters["a"],
        scattering_factor_b=scattering_factor_parameters["b"],
    )
    ```

    Alternatively, use the following to load with B-factors:

    ```python
    from cryojax.io import read_atoms_from_pdb
    from cryojax.simulator import PengAtomicPotential

    # Load positions of atoms, encoded atom names, and B-factors
    filename = "example.pdb"
    atom_positions, atom_identities, b_factors = read_atoms_from_pdb(
        filename, get_b_factors=True
    )
    scattering_factor_parameters = get_tabulated_scattering_factor_parameters(
        atom_identities, read_peng_element_scattering_factor_parameter_table()
        )
    potential = PengAtomicPotential(
        atom_positions,
        scattering_factor_a=scattering_factor_parameters["a"],
        scattering_factor_b=scattering_factor_parameters["b"],
        b_factors=b_factors,
    )
    ```

    **References:**

    - Peng, L-M. "Electron atomic scattering factors and scattering potentials of crystals."
      Micron 30.6 (1999): 625-648.
    - Peng, L-M., et al. "Robust parameterization of elastic and absorptive electron atomic
      scattering factors." Acta Crystallographica Section A: Foundations of Crystallography
      52.2 (1996): 257-276.
    """  # noqa: E501

    atom_positions: Float[Array, "n_atoms 3"]
    scattering_factor_a: Float[Array, "n_atoms 5"]
    scattering_factor_b: Float[Array, "n_atoms 5"]
    b_factors: Optional[Float[Array, " n_atoms"]]

    def __init__(
        self,
        atom_positions: Float[Array, "n_atoms 3"] | Float[np.ndarray, "n_atoms 3"],
        scattering_factor_a: Float[Array, "n_atoms 5"] | Float[np.ndarray, "n_atoms 5"],
        scattering_factor_b: Float[Array, "n_atoms 5"] | Float[np.ndarray, "n_atoms 5"],
        b_factors: Optional[
            Float[Array, " n_atoms"] | Float[np.ndarray, " n_atoms"]
        ] = None,
    ):
        """**Arguments:**

        - `atom_positions`:
            The coordinates of the atoms in units of angstroms.
        - `scattering_factor_a`:
            Scattering factor parameters $\\{a_i\\}_{i = 1}^5$
            for each atom from Peng et al. (1996).
        - `scattering_factor_b`:
            Scattering factor parameters $\\{b_i\\}_{i = 1}^5$
            for each atom from Peng et al. (1996).
        - `b_factors`:
            The B-factors applied to each atom.
        """
        self.atom_positions = jnp.asarray(atom_positions)
        self.scattering_factor_a = jnp.asarray(scattering_factor_a)
        self.scattering_factor_b = jnp.asarray(scattering_factor_b)
        if b_factors is None:
            self.b_factors = None
        else:
            self.b_factors = error_if_negative(jnp.asarray(b_factors))

    @override
    def as_real_voxel_grid(
        self,
        shape: tuple[int, int, int],
        voxel_size: Float[Array, ""] | float,
        *,
        batch_size_for_z_planes: int = 1,
        n_batches_of_atoms: int = 1,
    ) -> Float[Array, "{shape[0]} {shape[1]} {shape[2]}"]:
        """Return a voxel grid of the potential in real space.

        Through the work of Peng et al. (1996), tabulated elastic electron scattering factors
        are defined as

        $$f^{(e)}(\\mathbf{q}) = \\sum\\limits_{i = 1}^5 a_i \\exp(- b_i |\\mathbf{q}|^2),$$

        where $a_i$ is stored as `PengAtomicPotential.scattering_factor_a` and $b_i$ is
        stored as `PengAtomicPotential.scattering_factor_b` for the scattering vector $\\mathbf{q}$.
        Under usual scattering approximations (i.e. the first-born approximation),
        the rescaled electrostatic potential energy $U(\\mathbf{r})$ is then given by
        $4 \\pi \\mathcal{F}^{-1}[f^{(e)}(\\boldsymbol{\\xi} / 2)](\\mathbf{r})$, which is computed
        analytically as

        $$U(\\mathbf{r}) = 4 \\pi \\sum\\limits_{i = 1}^5 \\frac{a_i}{(2\\pi (b_i / 8 \\pi^2))^{3/2}} \\exp(- \\frac{|\\mathbf{r} - \\mathbf{r}'|^2}{2 (b_i / 8 \\pi^2)}),$$

        where $\\mathbf{r}'$ is the position of the atom. Including an additional B-factor (denoted by
        $B$ and stored as `PengAtomicPotential.b_factors`) gives the expression for the potential
        $U(\\mathbf{r})$ of a single atom type and its fourier transform pair $\\tilde{U}(\\boldsymbol{\\xi}) \\equiv \\mathcal{F}[U](\\boldsymbol{\\xi})$,

        $$U(\\mathbf{r}) = 4 \\pi \\sum\\limits_{i = 1}^5 \\frac{a_i}{(2\\pi ((b_i + B) / 8 \\pi^2))^{3/2}} \\exp(- \\frac{|\\mathbf{r} - \\mathbf{r}'|^2}{2 ((b_i + B) / 8 \\pi^2)}),$$

        $$\\tilde{U}(\\boldsymbol{\\xi}) = 4 \\pi \\sum\\limits_{i = 1}^5 a_i \\exp(- (b_i + B) |\\boldsymbol{\\xi}|^2 / 4) \\exp(2 \\pi i \\boldsymbol{\\xi}\\cdot\\mathbf{r}'),$$

        where $\\mathbf{q} = \\boldsymbol{\\xi} / 2$ gives the relationship between the wave vector and the
        scattering vector.

        In practice, for a discretization on a grid with voxel size $\\Delta r$ and grid point $\\mathbf{r}_{\\ell}$,
        the potential is evaluated as the average value inside the voxel

        $$U_{\\ell} = 4 \\pi \\frac{1}{\\Delta r^3} \\sum\\limits_{i = 1}^5 a_i \\prod\\limits_{j = 1}^3 \\int_{r^{\\ell}_j-\\Delta r/2}^{r^{\\ell}_j+\\Delta r/2} dr_j \\ \\frac{1}{{\\sqrt{2\\pi ((b_i + B) / 8 \\pi^2)}}} \\exp(- \\frac{(r_j - r'_j)^2}{2 ((b_i + B) / 8 \\pi^2)}),$$

        where $j$ indexes the components of the spatial coordinate vector $\\mathbf{r}$. The above expression is evaluated using the error function as

        $$U_{\\ell} = 4 \\pi \\frac{1}{(2 \\Delta r)^3} \\sum\\limits_{i = 1}^5 a_i \\prod\\limits_{j = 1}^3 \\textrm{erf}(\\frac{r_j^{\\ell} - r'_j + \\Delta r / 2}{\\sqrt{2 ((b_i + B) / 8\\pi^2)}}) - \\textrm{erf}(\\frac{r_j^{\\ell} - r'_j - \\Delta r / 2}{\\sqrt{2 ((b_i + B) / 8\\pi^2)}}).$$

        **Arguments:**

        - `shape`: The shape of the resulting voxel grid.
        - `voxel_size`: The voxel size of the resulting voxel grid.
        - `batch_size_for_z_planes`:
            The number of z-planes to evaluate in parallel with
            `jax.vmap`. By default, `1`.
        - `n_batches_of_atoms`:
            The number of iterations used to evaluate the volume,
            where the iteration is taken over groups of atoms.
            This is useful if `batch_size_for_z_planes = 1`
            and GPU memory is exhausted. By default, `1`.

        **Returns:**

        The rescaled potential $U_{\\ell}$ as a voxel grid of shape `shape`
        and voxel size `voxel_size`.
        """  # noqa: E501
        gaussian_amplitudes = self.scattering_factor_a
        if self.b_factors is None:
            gaussian_widths = self.scattering_factor_b
        else:
            gaussian_widths = self.scattering_factor_b + self.b_factors[:, None]
        return _build_real_space_voxel_potential_from_atoms(
            shape,
            jnp.asarray(voxel_size),
            self.atom_positions,
            gaussian_amplitudes,
            gaussian_widths,
            batch_size_for_z_planes=batch_size_for_z_planes,
            n_batches_of_atoms=n_batches_of_atoms,
        )


def _build_real_space_voxel_potential_from_atoms(
    shape: tuple[int, int, int],
    voxel_size: Float[Array, ""],
    atom_positions: Float[Array, "n_atoms 3"],
    a: Float[Array, "n_atoms n_gaussians_per_atom"],
    b: Float[Array, "n_atoms n_gaussians_per_atom"],
    batch_size_for_z_planes: int,
    n_batches_of_atoms: int,
) -> Float[Array, "{shape[0]} {shape[1]} {shape[2]}"]:
    # Make coordinate systems for each of x, y, and z dimensions
    z_dim, y_dim, x_dim = shape
    grid_x, grid_y, grid_z = [
        make_1d_coordinate_grid(dim, voxel_size) for dim in [x_dim, y_dim, z_dim]
    ]
    # Get function to compute potential over a batch of atoms
    compute_potential_for_atom_group = (
        lambda xs: _build_real_space_voxel_potential_from_atom_group(
            grid_x,
            grid_y,
            grid_z,
            voxel_size,
            xs[0],
            xs[1],
            xs[2],
            batch_size_for_z_planes,
        )
    )
    if n_batches_of_atoms > atom_positions.shape[0]:
        raise ValueError(
            "The `n_batches_of_atoms` when building a voxel grid must "
            "be an integer less than or equal to the number of atoms, "
            f"which is equal to {atom_positions.shape[0]}. Got "
            f"`n_batches_of_atoms = {n_batches_of_atoms}`."
        )
    elif n_batches_of_atoms == 1:
        potential_as_voxel_grid = compute_potential_for_atom_group((atom_positions, a, b))
    elif n_batches_of_atoms > 1:
        potential_as_voxel_grid = jnp.sum(
            _batched_map_with_n_batches(
                compute_potential_for_atom_group,
                (atom_positions, a, b),
                n_batches=n_batches_of_atoms,
                is_batch_axis_contracted=True,
            ),
            axis=0,
        )
    else:
        raise ValueError(
            "The `n_batches_of_atoms` when building a voxel grid must be an "
            "integer greater than or equal to 1."
        )

    return potential_as_voxel_grid


def _build_real_space_voxel_potential_from_atom_group(
    grid_x: Float[Array, " dim_x"],
    grid_y: Float[Array, " dim_y"],
    grid_z: Float[Array, " dim_z"],
    voxel_size: Float[Array, ""],
    atom_positions: Float[Array, "n_atoms_in_batch 3"],
    a: Float[Array, "n_atoms_in_batch n_gaussians_per_atom"],
    b: Float[Array, "n_atoms_in_batch n_gaussians_per_atom"],
    batch_size_for_z_planes: int,
) -> Float[Array, "dim_z dim_y dim_x"]:
    # Evaluate 1D gaussian integrals for each of x, y, and z dimensions
    (
        gaussian_integrals_times_prefactor_per_interval_per_atom_x,
        gaussian_integrals_per_interval_per_atom_y,
        gaussian_integrals_per_interval_per_atom_z,
    ) = _evaluate_gaussian_integrals_for_all_atoms_and_intervals(
        grid_x, grid_y, grid_z, atom_positions, a, b, voxel_size
    )
    # Get function to compute voxel grid at a single z-plane
    compute_potential_at_z_plane = (
        lambda gaussian_integrals_per_atom_z: _evaluate_gaussian_potential_at_z_plane(
            gaussian_integrals_times_prefactor_per_interval_per_atom_x,
            gaussian_integrals_per_interval_per_atom_y,
            gaussian_integrals_per_atom_z,
        )
    )
    # Map over z-planes
    if batch_size_for_z_planes > grid_z.size:
        raise ValueError(
            "The `batch_size_for_z_planes` when building a voxel grid must be an "
            "integer less than or equal to the z-dimension of the grid, "
            f"which is equal to {grid_z.size}."
        )
    elif batch_size_for_z_planes == 1:
        # ... compute the volume iteratively
        potential_as_voxel_grid = jax.lax.map(
            compute_potential_at_z_plane, gaussian_integrals_per_interval_per_atom_z
        )
    elif batch_size_for_z_planes > 1:
        # ... compute the volume by tuning how many z-planes to batch over
        compute_potential_at_z_planes = jax.vmap(compute_potential_at_z_plane, in_axes=0)
        potential_as_voxel_grid = _batched_map_with_batch_size(
            compute_potential_at_z_planes,
            gaussian_integrals_per_interval_per_atom_z,
            batch_size=batch_size_for_z_planes,
            is_batch_axis_contracted=False,
        )
    else:
        raise ValueError(
            "The `batch_size_for_z_planes` when building a voxel grid must be an "
            "integer greater than or equal to 1."
        )

    return potential_as_voxel_grid


def _evaluate_gaussian_integrals_for_all_atoms_and_intervals(
    grid_x: Float[Array, " dim_x"],
    grid_y: Float[Array, " dim_y"],
    grid_z: Float[Array, " dim_z"],
    atom_positions: Float[Array, "n_atoms 3"],
    a: Float[Array, "n_atoms n_gaussians_per_atom"],
    b: Float[Array, "n_atoms n_gaussians_per_atom"],
    voxel_size: Float[Array, ""],
) -> tuple[
    Float[Array, "dim_x n_atoms n_gaussians_per_atom"],
    Float[Array, "dim_y n_atoms n_gaussians_per_atom"],
    Float[Array, "dim_z n_atoms n_gaussians_per_atom"],
]:
    """Evaluate 1D averaged gaussians in x, y, and z dimensions
    for each atom and each gaussian per atom.
    """
    # Define function to compute integrals for each dimension
    scaling = 2 * jnp.pi / jnp.sqrt(b)
    integration_kernel = lambda delta: (
        jsp.special.erf(scaling[None, :, :] * (delta + voxel_size)[:, :, None])
        - jsp.special.erf(scaling[None, :, :] * delta[:, :, None])
    )
    # Compute outer product of left edge of grid points minus atomic positions
    left_edge_grid_x, left_edge_grid_y, left_edge_grid_z = (
        grid_x - voxel_size / 2,
        grid_y - voxel_size / 2,
        grid_z - voxel_size / 2,
    )
    delta_x, delta_y, delta_z = (
        left_edge_grid_x[:, None] - atom_positions[:, 0],
        left_edge_grid_y[:, None] - atom_positions[:, 1],
        left_edge_grid_z[:, None] - atom_positions[:, 2],
    )
    # Compute gaussian integrals for each grid point, each atom, and
    # each gaussian per atom
    gauss_x, gauss_y, gauss_z = (
        integration_kernel(delta_x),
        integration_kernel(delta_y),
        integration_kernel(delta_z),
    )
    # Compute the prefactors for each atom and each gaussian per atom
    # for the potential
    prefactor = (4 * jnp.pi * a) / (2 * voxel_size) ** 3
    # Multiply the prefactor onto one of the gaussians for efficiency
    return prefactor * gauss_x, gauss_y, gauss_z


def _evaluate_gaussian_potential_at_z_plane(
    gaussian_integrals_per_interval_per_atom_x: Float[
        Array, "dim_x n_atoms n_gaussians_per_atom"
    ],
    gaussian_integrals_per_interval_per_atom_y: Float[
        Array, "dim_y n_atoms n_gaussians_per_atom"
    ],
    gaussian_integrals_per_atom_z: Float[Array, "n_atoms n_gaussians_per_atom"],
) -> Float[Array, "dim_y dim_x"]:
    # Prepare matrices with dimensions of the number of atoms and the number of grid
    # points. There are as many matrices as number of gaussians per atom
    gauss_x = jnp.transpose(gaussian_integrals_per_interval_per_atom_x, (2, 1, 0))
    gauss_yz = jnp.transpose(
        gaussian_integrals_per_interval_per_atom_y
        * gaussian_integrals_per_atom_z[None, :, :],
        (2, 0, 1),
    )
    # Compute matrix multiplication then sum over the number of gaussians per atom
    return jnp.sum(jnp.matmul(gauss_yz, gauss_x), axis=0)


def _batched_map_with_n_batches(
    fun: Callable,
    xs: PyTree[Array],
    n_batches: int,
    is_batch_axis_contracted: bool = False,
):
    batch_dim = jtu.tree_leaves(xs)[0].shape[0]
    batch_size = batch_dim // n_batches
    return _batched_map(
        fun, xs, batch_dim, n_batches, batch_size, is_batch_axis_contracted
    )


def _batched_map_with_batch_size(
    fun: Callable,
    xs: PyTree[Array],
    batch_size: int,
    is_batch_axis_contracted: bool = False,
):
    batch_dim = jtu.tree_leaves(xs)[0].shape[0]
    n_batches = batch_dim // batch_size
    return _batched_map(
        fun, xs, batch_dim, n_batches, batch_size, is_batch_axis_contracted
    )


def _batched_map(
    fun: Callable,
    xs: PyTree[Array],
    batch_dim: int,
    n_batches: int,
    batch_size: int,
    is_batch_axis_contracted: bool = False,
):
    """Like `jax.lax.map`, but map over leading axis of `xs` in
    chunks of size `batch_size`. Assumes `fun` can be evaluated in
    parallel over this leading axis.
    """
    # ... reshape into an iterative dimension and a batching dimension
    xs_per_batch = jtu.tree_map(
        lambda x: x[: batch_dim - batch_dim % batch_size, ...].reshape(
            (n_batches, batch_size, *x.shape[1:])
        ),
        xs,
    )
    # .. compute the result and reshape back into one leading dimension
    result_per_batch = jax.lax.map(fun, xs_per_batch)
    if is_batch_axis_contracted:
        result = result_per_batch
    else:
        result = result_per_batch.reshape(
            (n_batches * batch_size, *result_per_batch.shape[2:])
        )
    # ... if the batch dimension is not divisible by the batch size, need
    # to take care of the remainder
    if batch_dim % batch_size != 0:
        remainder = fun(
            jtu.tree_map(lambda x: x[batch_dim - batch_dim % batch_size :, ...], xs)
        )
        if is_batch_axis_contracted:
            remainder = remainder[None, ...]
        result = jnp.concatenate([result, remainder], axis=0)
    return result



from typing import Callable, TypeVar

import equinox as eqx
import jax
from jaxtyping import Array, PyTree, Shaped


X = TypeVar("X")
Y = TypeVar("Y")
Carry = TypeVar("Carry")


def batched_map(
    f: Callable[
        [PyTree[Shaped[Array, "_ ..."], "X"]], PyTree[Shaped[Array, "_ ..."], "Y"]
    ],
    xs: PyTree[Shaped[Array, "_ ..."], "X"],
    *,
    batch_size: int = 1,
) -> PyTree[Shaped[Array, "_ ..."], "Y"]:
    """Like `jax.lax.map(..., batch_size=...)`, except
    `f(x)` is already vmapped by the user. In particular,
    it must be vmapped over the first axis of the arrays of `x`.

    **Arguments:**

    - `f`:
        As `jax.lax.map` with format `f(x)`, except
        vmapped over the first axis of the arrays of `x`.
    - `xs`:
        As `jax.lax.map`.
    - `batch_size`:
        Compute a loop of vmaps over `xs` in chunks of `batch_size`.

    **Returns:**

    As `jax.lax.map`.
    """

    @eqx.filter_jit
    def f_scan(carry, x):
        return carry, f(x)

    _, ys = batched_scan(f_scan, None, xs, batch_size=batch_size)

    return ys


def batched_scan(
    f: Callable[
        [Carry, PyTree[Shaped[Array, "_ ..."], "X"]],
        tuple[Carry, PyTree[Shaped[Array, "_ ..."], "Y"]],
    ],
    init: Carry,
    xs: PyTree[Shaped[Array, "_ ..."], "X"],
    length: int | None = None,
    unroll: int | bool = 1,
    *,
    batch_size: int = 1,
) -> tuple[Carry, PyTree[Shaped[Array, "_ ..."], "Y"]]:
    """Like `jax.lax.map(..., batch_size=...)`, except adding
    a `batch_size` to `jax.lax.scan`. Additionally, unlike
    `jax.lax.map`, it is assumed that `f(carry, x)` is already
    vmapped over the first axis of the arrays of `x`.

    **Arguments:**

    - `f`:
        As `jax.lax.scan` with format `f(carry, x)`, except
        vmapped over the first axis of the arrays of `x`.
    - `init`:
        As `jax.lax.scan`.
    - `xs`:
        As `jax.lax.scan`.
    - `length`:
        As `jax.lax.scan`.
    - `unroll`:
        As `jax.lax.scan`.
    - `batch_size`:
        Compute a loop of vmaps over `xs` in chunks of `batch_size`.

    **Returns:**

    As `jax.lax.scan`.
    """
    batch_dim = jax.tree.leaves(xs)[0].shape[0]
    n_batches = batch_dim // batch_size
    # Scan over batches
    scan_xs = jax.tree.map(
        lambda x: x[: batch_dim - batch_dim % batch_size, ...].reshape(
            (n_batches, batch_size, *x.shape[1:])
        ),
        xs,
    )
    carry, scan_ys = jax.lax.scan(f, init, scan_xs, length=length, unroll=unroll)
    ys = jax.tree.map(lambda y: y.reshape(n_batches * batch_size, *y.shape[2:]), scan_ys)
    if batch_dim % batch_size != 0:
        remainder_xs = jax.tree.map(
            lambda x: x[batch_dim - batch_dim % batch_size :, ...], xs
        )
        carry, remainder_ys = f(carry, remainder_xs)
        ys = jax.tree.map(
            lambda x, y: jax.lax.concatenate([x, y], dimension=0),
            ys,
            remainder_ys,
        )

    return carry, ys



from .base_image_model import (
    AbstractImageModel as AbstractImageModel,
    LinearImageModel as LinearImageModel,
    ProjectionImageModel as ProjectionImageModel,
)
from .physical_image_model import (
    AbstractPhysicalImageModel as AbstractPhysicalImageModel,
    ContrastImageModel as ContrastImageModel,
    ElectronCountsImageModel as ElectronCountsImageModel,
    IntensityImageModel as IntensityImageModel,
)



"""
Abstractions of ensembles of biological specimen.
"""

from abc import abstractmethod
from typing_extensions import override

import equinox as eqx

from .._pose import AbstractPose, EulerAnglePose
from .._potential_representation import (
    AbstractAtomicPotential,
    AbstractPotentialRepresentation,
)


class AbstractStructure(eqx.Module, strict=True):
    """A map from a pose to an `AbstractPotentialRepresentation`."""

    pose: eqx.AbstractVar[AbstractPose]

    @abstractmethod
    def get_potential_in_body_frame(self) -> AbstractPotentialRepresentation:
        """Get the scattering potential representation."""
        raise NotImplementedError

    def get_potential_in_transformed_frame(
        self, *, apply_translation: bool = False, apply_inverse_rotation: bool = False
    ) -> AbstractPotentialRepresentation:
        """Get the scattering potential, transformed by the pose."""
        potential = self.get_potential_in_body_frame()
        transformed_potential = potential.rotate_to_pose(
            self.pose, inverse=apply_inverse_rotation
        )
        if isinstance(transformed_potential, AbstractAtomicPotential):
            if apply_translation:
                transformed_potential = transformed_potential.translate_to_pose(self.pose)
        return transformed_potential


class BasicStructure(AbstractStructure, strict=True):
    """An "ensemble" with one conformation."""

    potential: AbstractPotentialRepresentation
    pose: AbstractPose

    def __init__(self, potential: AbstractPotentialRepresentation, pose: AbstractPose):
        """**Arguments:**

        - `potential`:
            The scattering potential representation of the specimen.
        - `pose`:
            The pose of the specimen.
        """
        self.potential = potential
        self.pose = pose or EulerAnglePose()

    @override
    def get_potential_in_body_frame(self) -> AbstractPotentialRepresentation:
        """Get the scattering potential in the center of mass
        frame.
        """
        return self.potential



"""
Routines for loading atomic structures.
Large amounts of the code are adapted from the ioSPI package
"""

import importlib.resources as pkg_resources
import os
from typing import Optional

import jax.numpy as jnp
import numpy as np
import xarray as xr
from jaxtyping import Float, Int


def get_tabulated_scattering_factor_parameters(
    atom_identities: Int[np.ndarray, " n_atoms"],
    scattering_factor_parameter_table: Optional[xr.Dataset] = None,
) -> dict[str, Float[np.ndarray, " n_atoms n_scattering_factors"]]:
    """Gets the parameters for the scattering factor for each atom in
    `atom_identities`.

    **Arguments:**

    - `atom_identitites`:
        Array containing the index of the one-hot encoded atom names.
        By default, Hydrogen is "1", Carbon is "6", Nitrogen is "7", etc.
    - `scattering_factor_parameter_table`:
        The table of scattering factors as an `xarray.Dataset`. By default, this
        is the tabulation from "Robust Parameterization of Elastic and
        Absorptive Electron Atomic Scattering Factors" by Peng et al. (1996),
        given by [`read_peng_element_scattering_factor_parameter_table`](https://mjo22.github.io/cryojax/api/constants/scattering_factor_parameters/#cryojax.constants.read_peng_element_scattering_factor_parameter_table).

    **Returns:**

    The particular scattering factor parameters stored in
    `scattering_factor_parameter_table` for `atom_identities`.
    """  # noqa: E501
    if scattering_factor_parameter_table is None:
        scattering_factor_parameter_table = (
            read_peng_element_scattering_factor_parameter_table()
        )
    return {
        str(k): np.asarray(v.data[np.asarray(atom_identities), ...])
        for k, v in scattering_factor_parameter_table.items()
    }


def read_peng_element_scattering_factor_parameter_table() -> xr.Dataset:
    r"""Function to load the atomic scattering factor parameter
    table from "Robust Parameterization of Elastic and Absorptive
    Electron Atomic Scattering Factors" by Peng et al. (1996).

    **Returns:**

    The parameter table for parameters $\{a_i\}_{i = 1}^5$ and $\{b_i\}_{i = 1}^5$
    for each atom, described in the above reference. This is stored as an
    `xarray.Dataset`.
    """
    with pkg_resources.as_file(
        pkg_resources.files("cryojax").joinpath("constants")
    ) as path:
        scattering_factor_parameters = jnp.load(
            os.path.join(path, "peng1996_element_params.npy")
        )
    dims = ("n_elements", "n_gaussians")
    scattering_factor_parameter_table = xr.Dataset(
        data_vars=dict(
            a=xr.DataArray(scattering_factor_parameters[0], dims=dims),
            b=xr.DataArray(scattering_factor_parameters[1], dims=dims),
        ),
        attrs=dict(description="Scattering factor parameters from Peng et al. (1996)"),
    )

    return scattering_factor_parameter_table



"""
Image formation models.
"""

from typing import Optional
from typing_extensions import override

import equinox as eqx
import jax
from jaxtyping import Array, Bool, PRNGKeyArray

from .._config import AbstractConfig, DoseConfig
from .._detector import AbstractDetector
from .._scattering_theory import AbstractScatteringTheory
from .._structure import AbstractStructure
from .base_image_model import AbstractImageModel, ImageArray, PaddedImageArray


class AbstractPhysicalImageModel(AbstractImageModel, strict=True):
    """An image formation model that simulates physical
    quantities. This uses the `AbstractScatteringTheory` class.
    """

    scattering_theory: eqx.AbstractVar[AbstractScatteringTheory]


class ContrastImageModel(AbstractPhysicalImageModel, strict=True):
    """An image formation model that returns the image contrast from a linear
    scattering theory.
    """

    structure: AbstractStructure
    config: AbstractConfig
    scattering_theory: AbstractScatteringTheory

    normalizes_signal: bool
    signal_region: Optional[Bool[Array, "_ _"]]

    def __init__(
        self,
        structure: AbstractStructure,
        config: AbstractConfig,
        scattering_theory: AbstractScatteringTheory,
        *,
        normalizes_signal: bool = False,
        signal_region: Optional[Bool[Array, "_ _"]] = None,
    ):
        """**Arguments:**

        - `structure`:
            The biological structure.
        - `config`:
            The configuration of the instrument, such as for the pixel size
            and the wavelength.
        - `scattering_theory`:
            The scattering theory.
        - `normalizes_signal`:
            If `True`, normalize the image before returning.
        - `signal_region`:
            A boolean array that is 1 where there is signal,
            and 0 otherwise used to normalize the image.
            Must have shape equal to `AbstractImageModel.shape`.
        """
        self.structure = structure
        self.config = config
        self.scattering_theory = scattering_theory
        self.normalizes_signal = normalizes_signal
        self.signal_region = signal_region

    @override
    def compute_fourier_image(
        self, rng_key: Optional[PRNGKeyArray] = None
    ) -> ImageArray | PaddedImageArray:
        # Get the potential
        potential = self.structure.get_potential_in_transformed_frame(
            apply_inverse_rotation=self.scattering_theory.integrator.requires_inverse_rotation
        )
        # Compute the squared wavefunction
        contrast_spectrum = self.scattering_theory.compute_contrast_spectrum(
            potential,
            self.config,
            rng_key,
            defocus_offset=self.structure.pose.offset_z_in_angstroms,
        )
        # Apply the translation
        contrast_spectrum = self._apply_translation(contrast_spectrum)

        return contrast_spectrum


class IntensityImageModel(AbstractPhysicalImageModel, strict=True):
    """An image formation model that returns an intensity distribution---or in other
    words a squared wavefunction.
    """

    structure: AbstractStructure
    config: AbstractConfig
    scattering_theory: AbstractScatteringTheory

    normalizes_signal: bool
    signal_region: Optional[Bool[Array, "_ _"]]

    def __init__(
        self,
        structure: AbstractStructure,
        config: AbstractConfig,
        scattering_theory: AbstractScatteringTheory,
        *,
        normalizes_signal: bool = False,
        signal_region: Optional[Bool[Array, "_ _"]] = None,
    ):
        """**Arguments:**

        - `structure`:
            The biological structure.
        - `config`:
            The configuration of the instrument, such as for the pixel size
            and the wavelength.
        - `scattering_theory`:
            The scattering theory.
        - `normalizes_signal`:
            If `True`, normalize the image before returning.
        - `signal_region`:
            A boolean array that is 1 where there is signal,
            and 0 otherwise used to normalize the image.
            Must have shape equal to `AbstractImageModel.shape`.
        """
        self.structure = structure
        self.config = config
        self.scattering_theory = scattering_theory
        self.normalizes_signal = normalizes_signal
        self.signal_region = signal_region

    @override
    def compute_fourier_image(
        self, rng_key: Optional[PRNGKeyArray] = None
    ) -> ImageArray | PaddedImageArray:
        potential = self.structure.get_potential_in_transformed_frame(
            apply_inverse_rotation=self.scattering_theory.integrator.requires_inverse_rotation
        )
        scattering_theory = self.scattering_theory
        fourier_intensity = scattering_theory.compute_intensity_spectrum(
            potential,
            self.config,
            rng_key,
            defocus_offset=self.structure.pose.offset_z_in_angstroms,
        )
        fourier_intensity = self._apply_translation(fourier_intensity)

        return fourier_intensity


class ElectronCountsImageModel(AbstractPhysicalImageModel, strict=True):
    """An image formation model that returns electron counts, given a
    model for the detector.
    """

    structure: AbstractStructure
    config: DoseConfig
    scattering_theory: AbstractScatteringTheory
    detector: AbstractDetector

    normalizes_signal: bool
    signal_region: Optional[Bool[Array, "_ _"]]

    def __init__(
        self,
        structure: AbstractStructure,
        config: DoseConfig,
        scattering_theory: AbstractScatteringTheory,
        detector: AbstractDetector,
        *,
        normalizes_signal: bool = False,
        signal_region: Optional[Bool[Array, "_ _"]] = None,
    ):
        """**Arguments:**

        - `structure`:
            The biological structure.
        - `config`:
            The configuration of the instrument, such as for the pixel size
            and the wavelength.
        - `scattering_theory`:
            The scattering theory.
        - `normalizes_signal`:
            If `True`, normalize the image before returning.
        - `signal_region`:
            A boolean array that is 1 where there is signal,
            and 0 otherwise used to normalize the image.
            Must have shape equal to `AbstractImageModel.shape`.
        """
        self.structure = structure
        self.config = config
        self.scattering_theory = scattering_theory
        self.detector = detector
        self.normalizes_signal = normalizes_signal
        self.signal_region = signal_region

    @override
    def compute_fourier_image(
        self, rng_key: Optional[PRNGKeyArray] = None
    ) -> ImageArray | PaddedImageArray:
        potential = self.structure.get_potential_in_transformed_frame(
            apply_inverse_rotation=self.scattering_theory.integrator.requires_inverse_rotation
        )
        if rng_key is None:
            # Compute the squared wavefunction
            scattering_theory = self.scattering_theory
            fourier_intensity = scattering_theory.compute_intensity_spectrum(
                potential,
                self.config,
                defocus_offset=self.structure.pose.offset_z_in_angstroms,
            )
            fourier_intensity = self._apply_translation(fourier_intensity)
            # ... now measure the expected electron events at the detector
            fourier_expected_electron_events = (
                self.detector.compute_expected_electron_events(
                    fourier_intensity, self.config
                )
            )

            return fourier_expected_electron_events
        else:
            keys = jax.random.split(rng_key)
            # Compute the squared wavefunction
            scattering_theory = self.scattering_theory
            fourier_intensity = scattering_theory.compute_intensity_spectrum(
                potential,
                self.config,
                keys[0],
                defocus_offset=self.structure.pose.offset_z_in_angstroms,
            )
            fourier_intensity = self._apply_translation(fourier_intensity)
            # ... now measure the detector readout
            fourier_detector_readout = self.detector.compute_detector_readout(
                keys[1],
                fourier_intensity,
                self.config,
            )

            return fourier_detector_readout



from ..simulator._potential_integrator import (
    AbstractMultisliceIntegrator as AbstractMultisliceIntegrator,
    EwaldSphereExtraction as EwaldSphereExtraction,
    FFTMultisliceIntegrator as FFTMultisliceIntegrator,
)
from ..simulator._scattering_theory import (
    AbstractWaveScatteringTheory as AbstractWaveScatteringTheory,
    HighEnergyScatteringTheory as HighEnergyScatteringTheory,
    MultisliceScatteringTheory as MultisliceScatteringTheory,
)
from ..simulator._solvent import (
    GRFSolvent as GRFSolvent,
    SolventMixturePower as SolventMixturePower,
)
from ..simulator._transfer_theory import (
    WaveTransferTheory as WaveTransferTheory,
)



import numpy as np
import pytest

import cryojax.simulator as cxs
from cryojax.io import read_array_from_mrc


@pytest.fixture
def voxel_potential(sample_mrc_path):
    real_voxel_grid, voxel_size = read_array_from_mrc(sample_mrc_path, loads_spacing=True)
    return cxs.FourierVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size, pad_scale=1.3
    )


@pytest.fixture
def basic_config(voxel_potential):
    return cxs.BasicConfig(
        shape=voxel_potential.shape[0:2],
        pixel_size=voxel_potential.voxel_size,
        voltage_in_kilovolts=300.0,
    )


@pytest.fixture
def image_model(voxel_potential, basic_config):
    image_model = cxs.make_image_model(
        voxel_potential,
        basic_config,
        pose=cxs.EulerAnglePose(),
        transfer_theory=cxs.ContrastTransferTheory(cxs.AberratedAstigmaticCTF()),
    )
    return image_model


@pytest.mark.parametrize(
    "cls, model",
    [
        (cxs.IndependentGaussianPixels, "image_model"),
        (cxs.IndependentGaussianFourierModes, "image_model"),
    ],
)
def test_simulate_signal_from_gaussian_distributions(cls, model, request):
    model = request.getfixturevalue(model)
    distribution = cls(model)
    np.testing.assert_allclose(model.simulate(), distribution.compute_signal())



"""This is a module of utilities for working with JAX/Equinox. This includes utilities
for Equinox filtered transformations and Equinox recommendations for creating custom
per-leaf behavior for pytrees.
"""

from ._batched_loop import batched_map as batched_map, batched_scan as batched_scan
from ._filter_specs import get_filter_spec as get_filter_spec
from ._grid_search import (
    AbstractGridSearchMethod as AbstractGridSearchMethod,
    MinimumSearchMethod as MinimumSearchMethod,
    MinimumSolution as MinimumSolution,
    MinimumState as MinimumState,
    run_grid_search as run_grid_search,
    tree_grid_shape as tree_grid_shape,
    tree_grid_take as tree_grid_take,
    tree_grid_unravel_index as tree_grid_unravel_index,
)
from ._pytree_transforms import (
    AbstractPyTreeTransform as AbstractPyTreeTransform,
    CustomTransform as CustomTransform,
    StopGradientTransform as StopGradientTransform,
    resolve_transforms as resolve_transforms,
)



"""Unit conversions."""

import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, Float


def convert_keV_to_angstroms(
    energy_in_keV: Float[Array | np.ndarray, ""] | float,
) -> Float[Array, ""]:
    """Get the relativistic electron wavelength at a given accelerating voltage. For
    reference, see Equation 2.5 in Section 2.1 from *Spence, John CH. High-resolution
    electron microscopy. OUP Oxford, 2013.*.

    **Arguments:**

    - `energy_in_keV`:
        The energy in kiloelectron volts.

    **Returns:**

    The relativistically corrected electron wavelength in Angstroms corresponding to the
    energy `energy_in_keV`.
    """
    energy_in_eV = 1000.0 * energy_in_keV  # keV to eV
    return jnp.asarray(12.2639 / (energy_in_eV + 0.97845e-6 * energy_in_eV**2) ** 0.5)



"""
Abstractions of ensembles of biological specimen.
"""

from typing import Any, Optional

import equinox as eqx

from .base_structure import AbstractStructure


class AbstractStructuralEnsemble(AbstractStructure, strict=True):
    """A map from a pose and conformational variable to an
    `AbstractPotentialRepresentation`.
    """

    conformation: eqx.AbstractVar[Optional[Any]]



from ._conventions import (
    convert_b_factor_to_variance as convert_b_factor_to_variance,
    convert_variance_to_b_factor as convert_variance_to_b_factor,
)
from ._parkhurst2024_solvent_power import (
    PARKHURST2024_POWER_CONSTANTS as PARKHURST2024_POWER_CONSTANTS,
)
from ._scattering_factor_parameters import (
    get_tabulated_scattering_factor_parameters as get_tabulated_scattering_factor_parameters,  # noqa: E501
    read_peng_element_scattering_factor_parameter_table as read_peng_element_scattering_factor_parameter_table,  # noqa: E501
)
from ._unit_conversions import (
    convert_keV_to_angstroms as convert_keV_to_angstroms,
)



from .base_multislice_integrator import (
    AbstractMultisliceIntegrator as AbstractMultisliceIntegrator,
)
from .fft_multislice_integrator import FFTMultisliceIntegrator as FFTMultisliceIntegrator



"""
Representations of rigid body rotations and translations of 3D coordinate systems.
"""

from abc import abstractmethod
from functools import cached_property
from typing import Optional
from typing_extensions import Self, override

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
from equinox import AbstractVar, Module
from jaxtyping import Array, Complex, Float

from ..ndimage import enforce_self_conjugate_rfftn_components
from ..rotations import SO3, convert_quaternion_to_euler_angles


class AbstractPose(Module, strict=True):
    """Base class for the image pose. Subclasses will choose a
    particular convention for parameterizing the rotation by
    overwriting the `AbstractPose.rotation` property.
    """

    offset_x_in_angstroms: AbstractVar[Float[Array, ""]]
    offset_y_in_angstroms: AbstractVar[Float[Array, ""]]
    offset_z_in_angstroms: AbstractVar[Optional[Float[Array, ""]]]

    def rotate_coordinates(
        self,
        coordinate_grid_or_list: (
            Float[Array, "z_dim y_dim x_dim 3"] | Float[Array, "size 3"]
        ),
        inverse: bool = False,
    ) -> Float[Array, "z_dim y_dim x_dim 3"] | Float[Array, "size 3"]:
        """Rotate a 3D coordinate system.

        **Arguments:**

        - `coordinate_grid_or_list`:
            The 3D coordinate system to rotate. This can either be a list of coordinates
            of shape `(N, 3)` or a grid of coordinates `(N1, N2, N3, 3)`.
        - `inverse`:
            If `True`, compute the inverse rotation (i.e. rotation by the matrix $R^T$,
            where $R$ is the rotation matrix).

        **Returns:**

        The rotated version of `coordinate_grid_or_list`.
        """
        rotation = self.rotation.inverse() if inverse else self.rotation
        if isinstance(coordinate_grid_or_list, Float[Array, "size 3"]):  # type: ignore
            rotated_coordinate_grid_or_list = jax.vmap(rotation.apply)(
                coordinate_grid_or_list
            )
        elif isinstance(coordinate_grid_or_list, Float[Array, "z_dim y_dim x_dim 3"]):  # type: ignore
            rotated_coordinate_grid_or_list = jax.vmap(
                jax.vmap(jax.vmap(rotation.apply))
            )(coordinate_grid_or_list)
        else:
            raise ValueError(
                "Coordinates must be a JAX array either of shape (N, 3) or "
                f"(N1, N2, N3, 3). Instead, got {coordinate_grid_or_list.shape} and type "
                f"{type(coordinate_grid_or_list)}."
            )
        return rotated_coordinate_grid_or_list

    def translate_image(
        self,
        fourier_image: Complex[Array, "{shape[0]} {shape[1]}//2+1"],
        translation_operator: Complex[Array, "{shape[0]} {shape[1]}//2+1"],
        shape: tuple[int, int],
    ) -> Complex[Array, "{shape[0]} {shape[1]}//2+1"]:
        """Apply translational phase shifts to a fourier-space image.

        **Arguments:**

        - `fourier_image`:
            The image in fourier-space, which is the output of a call
            to `cryojax.image.rfftn`.
        - `phase_shifts`:
            The phase shifts for translation, which are computed from
            `AbstractPose.compute_translation_operator`.
        - `shape`:
            The shape of `fourier_image` in real-space.

        **Return:**

        The translated `fourier_image`, taking care to avoid image
        artifacts when applying the phase shifts.
        """
        fourier_image = enforce_self_conjugate_rfftn_components(
            fourier_image, shape, includes_zero_frequency=False, mode="zero"
        )
        return fourier_image * translation_operator

    def compute_translation_operator(
        self, frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"]
    ) -> Complex[Array, "y_dim x_dim"]:
        """Compute the phase shifts from the in-plane translation,
        given a frequency grid coordinate system.

        **Arguments:**

        - `frequency_grid_in_angstroms`:
            A grid of in-plane frequency coordinates $(q_x, q_y)$

        **Returns:**

        From the vector $(t_x, t_y)$ (given by `self.offset_in_angstroms`), returns the
        grid of in-plane phase shifts $\\exp{(- 2 \\pi i (t_x q_x + t_y q_y))}$.
        """
        xy = self.offset_in_angstroms[0:2]
        return jnp.exp(-1.0j * (2 * jnp.pi * jnp.matmul(frequency_grid_in_angstroms, xy)))

    @cached_property
    def offset_in_angstroms(self) -> Float[Array, "2"] | Float[Array, "3"]:
        """The in-plane translation vector, where the origin in taken to
        be in the center of the imaging plane.
        """
        if self.offset_z_in_angstroms is None:
            return jnp.asarray(
                (
                    self.offset_x_in_angstroms,
                    self.offset_y_in_angstroms,
                )
            )
        else:
            return jnp.asarray(
                (
                    self.offset_x_in_angstroms,
                    self.offset_y_in_angstroms,
                    self.offset_z_in_angstroms,
                )
            )

    @cached_property
    @abstractmethod
    def rotation(self) -> SO3:
        """Generate an `SO3` object from a particular angular
        parameterization.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_rotation(cls, rotation: SO3) -> Self:
        """Construct an `AbstractPose` from an `SO3` object."""
        raise NotImplementedError

    @classmethod
    def from_rotation_and_translation(
        cls,
        rotation: SO3,
        offset_in_angstroms: Float[Array, "2"] | Float[Array, "3"],
    ) -> Self:
        """Construct an `AbstractPose` from an `SO3` object and a
        translation vector.
        """
        if offset_in_angstroms.shape == (2,):
            return eqx.tree_at(
                lambda self: (
                    self.offset_x_in_angstroms,
                    self.offset_y_in_angstroms,
                ),
                cls.from_rotation(rotation),
                (
                    offset_in_angstroms[0],
                    offset_in_angstroms[1],
                ),
            )
        elif offset_in_angstroms.shape == (3,):
            return eqx.tree_at(
                lambda self: (
                    self.offset_x_in_angstroms,
                    self.offset_y_in_angstroms,
                    self.offset_z_in_angstroms,
                ),
                cls.from_rotation(rotation),
                (
                    offset_in_angstroms[0],
                    offset_in_angstroms[1],
                    offset_in_angstroms[2],
                ),
            )
        else:
            raise ValueError(
                "Array `offset_in_angstroms` given to constructor "
                f"`{cls.__name__}.from_rotation_and_translation` supports "
                "shapes `(2,)` and `(3,)`. Got shape "
                f"`{offset_in_angstroms.shape}`"
            )


class EulerAnglePose(AbstractPose, strict=True):
    r"""An `AbstractPose` represented by Euler angles.
    Angles are given in degrees, and the sequence of rotations is a
    zyz *extrinsic* rotation, with `phi_angle` as the first euler angle,
    `theta_angle` as the second, and `psi_angle` is the third.

    !!! info "Converting to RELION and FREALIGN convention"

        RELION/FREALIGN convention is that the euler angles represent
        a zyz *intrinsic* rotation that "undoes" the rotation in the image. cryoJAX
        defines its convention to be a zyz *extrinsic* rotation that generates the
        pose in the image. In order to convert to the RELION/FREALIGN convention,
        simply **negate each euler angle**.
    """

    offset_x_in_angstroms: Float[Array, ""]
    offset_y_in_angstroms: Float[Array, ""]
    offset_z_in_angstroms: Optional[Float[Array, ""]]

    phi_angle: Float[Array, ""]
    theta_angle: Float[Array, ""]
    psi_angle: Float[Array, ""]

    def __init__(
        self,
        offset_x_in_angstroms: float | Float[Array, ""] = 0.0,
        offset_y_in_angstroms: float | Float[Array, ""] = 0.0,
        phi_angle: float | Float[Array, ""] = 0.0,
        theta_angle: float | Float[Array, ""] = 0.0,
        psi_angle: float | Float[Array, ""] = 0.0,
        *,
        offset_z_in_angstroms: Optional[float | Float[Array, ""]] = None,
    ):
        """**Arguments:**

        - `offset_x_in_angstroms`: In-plane translation in x direction.
        - `offset_y_in_angstroms`: In-plane translation in y direction.
        - `phi_angle`: Angle to rotate about first rotation axis, which is the z axis.
        - `theta_angle`: Angle to rotate about second rotation axis, which is the y axis.
        - `psi_angle`: Angle to rotate about third rotation axis, which is the z axis.
        - `offset_z_in_angstroms`: Out-of-plane translation in z direction.
        """
        self.offset_x_in_angstroms = jnp.asarray(offset_x_in_angstroms, dtype=float)
        self.offset_y_in_angstroms = jnp.asarray(offset_y_in_angstroms, dtype=float)
        self.phi_angle = jnp.asarray(phi_angle, dtype=float)
        self.theta_angle = jnp.asarray(theta_angle, dtype=float)
        self.psi_angle = jnp.asarray(psi_angle, dtype=float)
        self.offset_z_in_angstroms = (
            None
            if offset_z_in_angstroms is None
            else jnp.asarray(offset_z_in_angstroms, dtype=float)
        )

    @cached_property
    @override
    def rotation(self) -> SO3:
        """Generate a `SO3` object from a set of Euler angles."""
        phi, theta, psi = self.phi_angle, self.theta_angle, self.psi_angle
        # Convert to radians.
        phi = jnp.deg2rad(phi)
        theta = jnp.deg2rad(theta)
        psi = jnp.deg2rad(psi)
        # Get sequence of rotations.
        R1, R2, R3 = (
            SO3.from_z_radians(phi),
            SO3.from_y_radians(theta),
            SO3.from_z_radians(psi),
        )
        return R3 @ R2 @ R1

    @override
    @classmethod
    def from_rotation(cls, rotation: SO3) -> Self:
        phi_angle, theta_angle, psi_angle = convert_quaternion_to_euler_angles(
            rotation.wxyz, convention="zyz", extrinsic=True
        )
        return cls(phi_angle=phi_angle, theta_angle=theta_angle, psi_angle=psi_angle)


class QuaternionPose(AbstractPose, strict=True):
    """An `AbstractPose` represented by unit quaternions."""

    offset_x_in_angstroms: Float[Array, ""]
    offset_y_in_angstroms: Float[Array, ""]
    offset_z_in_angstroms: Optional[Float[Array, ""]]

    wxyz: Float[Array, "4"]

    def __init__(
        self,
        offset_x_in_angstroms: float | Float[Array, ""] = 0.0,
        offset_y_in_angstroms: float | Float[Array, ""] = 0.0,
        wxyz: (
            tuple[float, float, float, float] | Float[np.ndarray, "4"] | Float[Array, "4"]
        ) = (1.0, 0.0, 0.0, 0.0),
        *,
        offset_z_in_angstroms: Optional[float | Float[Array, ""]] = None,
    ):
        """**Arguments:**

        - `offset_x_in_angstroms`: In-plane translation in x direction.
        - `offset_y_in_angstroms`: In-plane translation in y direction.
        - `wxyz`:
            The quaternion, represented as a vector $\\mathbf{q} = (q_w, q_x, q_y, q_z)$.
        - `offset_z_in_angstroms`: Out-of-plane translation in z direction.
        """
        self.offset_x_in_angstroms = jnp.asarray(offset_x_in_angstroms)
        self.offset_y_in_angstroms = jnp.asarray(offset_y_in_angstroms)
        self.wxyz = jnp.asarray(wxyz)
        self.offset_z_in_angstroms = (
            None
            if offset_z_in_angstroms is None
            else jnp.asarray(offset_z_in_angstroms, dtype=float)
        )

    @cached_property
    @override
    def rotation(self) -> SO3:
        """Generate rotation from the unit quaternion
        $\\mathbf{q} / |\\mathbf{q}|$.
        """
        # Generate SO3 object from unit quaternion
        R = SO3(wxyz=self.wxyz).normalize()
        return R

    @override
    @classmethod
    def from_rotation(cls, rotation: SO3) -> Self:
        return cls(wxyz=rotation.wxyz)


class AxisAnglePose(AbstractPose, strict=True):
    """An `AbstractPose` parameterized in the axis-angle representation.

    The axis-angle representation parameterizes elements of the so3 algebra,
    which are skew-symmetric matrices, with the euler vector
    $\\boldsymbol{\\omega} = (\\omega_x, \\omega_y, \\omega_z)$.
    The magnitude of this vector is the angle, and the unit vector is the axis.

    In a `SO3` object, the euler vector is mapped to SO3 group elements using
    the matrix exponential.
    """

    offset_x_in_angstroms: Float[Array, ""]
    offset_y_in_angstroms: Float[Array, ""]
    offset_z_in_angstroms: Optional[Float[Array, ""]]

    euler_vector: Float[Array, "3"]

    def __init__(
        self,
        offset_x_in_angstroms: float | Float[Array, ""] = 0.0,
        offset_y_in_angstroms: float | Float[Array, ""] = 0.0,
        euler_vector: (
            tuple[float, float, float] | Float[np.ndarray, "3"] | Float[Array, "3"]
        ) = (0.0, 0.0, 0.0),
        *,
        offset_z_in_angstroms: Optional[float | Float[Array, ""]] = None,
    ):
        """**Arguments:**

        - `offset_x_in_angstroms`: In-plane translation in x direction.
        - `offset_y_in_angstroms`: In-plane translation in y direction.
        - `euler_vector`:
            The axis-angle parameterization, represented with the euler
            vector $\\boldsymbol{\\omega}$.
        - `offset_z_in_angstroms`: Out-of-plane translation in z direction.
        """
        self.offset_x_in_angstroms = jnp.asarray(offset_x_in_angstroms)
        self.offset_y_in_angstroms = jnp.asarray(offset_y_in_angstroms)
        self.euler_vector = jnp.asarray(euler_vector)
        self.offset_z_in_angstroms = (
            None
            if offset_z_in_angstroms is None
            else jnp.asarray(offset_z_in_angstroms, dtype=float)
        )

    @cached_property
    @override
    def rotation(self) -> SO3:
        """Generate rotation from an euler vector using the exponential map."""
        # Convert degrees to radians
        euler_vector = jnp.deg2rad(self.euler_vector)
        # Project the tangent vector onto the manifold with
        # the exponential map
        R = SO3.exp(euler_vector)
        return R

    @override
    @classmethod
    def from_rotation(cls, rotation: SO3) -> Self:
        # Compute the euler vector from the logarithmic map
        euler_vector = jnp.rad2deg(rotation.log())
        return cls(euler_vector=euler_vector)



"""Routines for downsampling arrays"""

from typing import overload

import jax.numpy as jnp
from jaxtyping import Array, Complex, Float, Inexact

from ._edges import crop_to_shape
from ._fft import fftn, ifftn, rfftn


@overload
def downsample_with_fourier_cropping(
    image_or_volume: Inexact[Array, "_ _ _"],
    downsampling_factor: float | int,
    outputs_real_space: bool = True,
) -> Inexact[Array, "_ _ _"]: ...


@overload
def downsample_with_fourier_cropping(
    image_or_volume: Inexact[Array, "_ _"],
    downsampling_factor: float | int,
    outputs_real_space: bool = True,
) -> Inexact[Array, "_ _"]: ...


def downsample_with_fourier_cropping(
    image_or_volume: Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"],
    downsampling_factor: float | int,
    outputs_real_space: bool = True,
) -> Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]:
    """Downsample an array using fourier cropping.

    **Arguments:**

    - `image_or_volume`: The image or volume array to downsample.
    - `downsample_factor`:
        A scale factor at which to downsample `image_or_volume`
        by. Must be a value greater than `1`.
    - `outputs_real_space`:
        If `False`, the `image_or_volume` is returned in fourier space.

    **Returns:**

    The downsampled `image_or_volume`, at the new real-space shape
    `downsampled_shape`. If `outputs_real_space = False`, return
    the downsampled array in fourier space, with the zero frequency
    component in the corner. For real signals, hermitian symmetry is
    assumed.
    """
    downsampling_factor = float(downsampling_factor)
    if downsampling_factor < 1.0:
        raise ValueError("`downsampling_factor` must be greater than 1.0")
    if image_or_volume.ndim == 2:
        image = image_or_volume
        new_shape = (
            int(image.shape[0] / downsampling_factor),
            int(image.shape[1] / downsampling_factor),
        )
        downsampled_array = downsample_to_shape_with_fourier_cropping(
            image, new_shape, outputs_real_space=outputs_real_space
        )
    elif image_or_volume.ndim == 3:
        volume = image_or_volume
        new_shape = (
            int(volume.shape[0] / downsampling_factor),
            int(volume.shape[1] / downsampling_factor),
            int(volume.shape[2] / downsampling_factor),
        )
        downsampled_array = downsample_to_shape_with_fourier_cropping(
            volume, new_shape, outputs_real_space=outputs_real_space
        )
    else:
        raise ValueError(
            "`downsample_with_fourier_cropping` can only crop images and volumes. "
            f"Got an array with number of dimensions {image_or_volume.ndim}."
        )

    return downsampled_array


@overload
def downsample_to_shape_with_fourier_cropping(
    image_or_volume: Inexact[Array, "_ _"],
    downsampled_shape: tuple[int, int],
    outputs_real_space: bool = True,
) -> Inexact[Array, "_ _"]: ...


@overload
def downsample_to_shape_with_fourier_cropping(
    image_or_volume: Inexact[Array, "_ _ _"],
    downsampled_shape: tuple[int, int, int],
    outputs_real_space: bool = True,
) -> Inexact[Array, "_ _ _"]: ...


def downsample_to_shape_with_fourier_cropping(
    image_or_volume: Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"],
    downsampled_shape: tuple[int, int] | tuple[int, int, int],
    outputs_real_space: bool = True,
) -> Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]:
    """Downsample an array to a specified shape using fourier cropping.

    For real signals, the Hartley Transform is used to downsample the signal.
    For complex signals, the Fourier Transform is used to downsample the signal.

    The real case is based on the `downsample_transform` function in cryoDRGN
    https://github.com/ml-struct-bio/cryodrgn/blob/4ba75502d4dd1d0e5be3ecabf4a005c652edf4b5/cryodrgn/commands/downsample.py#L154

    **Arguments:**

    - `image_or_volume`: The image or volume array to downsample.
    - `downsampled_shape`:
        The new shape after fourier cropping.
    - `outputs_real_space`:
        If `False`, the `image_or_volume` is returned in fourier space.

    **Returns:**

    The downsampled `image_or_volume`, at the new real-space shape
    `downsampled_shape`. If `outputs_real_space = False`, return
    the downsampled array in fourier space, with the zero frequency
    component in the corner. For real signals, hermitian symmetry is
    assumed.
    """
    if jnp.iscomplexobj(image_or_volume):
        return _downsample_complex_signal_to_shape(
            image_or_volume, downsampled_shape, outputs_real_space=outputs_real_space
        )
    else:
        return _downsample_real_signal_to_shape(
            image_or_volume, downsampled_shape, outputs_real_space=outputs_real_space
        )


def _downsample_real_signal_to_shape(
    image_or_volume: Float[Array, "_ _"] | Float[Array, "_ _ _"],
    downsampled_shape: tuple[int, int] | tuple[int, int, int],
    outputs_real_space: bool = True,
) -> Inexact[Array, "_ _"] | Inexact[Array, "_ _ _"]:
    # Forward Hartley Transform
    hartley_array = jnp.fft.fftshift(fftn(image_or_volume))
    hartley_array = hartley_array.real - hartley_array.imag

    # Crop to the desired shape
    ds_image_or_volume = crop_to_shape(hartley_array, downsampled_shape)

    # Inverse Hartley Transform
    ds_image_or_volume = jnp.fft.fftshift(fftn(ds_image_or_volume))
    ds_image_or_volume /= ds_image_or_volume.size
    ds_image_or_volume = ds_image_or_volume.real - ds_image_or_volume.imag

    if outputs_real_space:
        return ds_image_or_volume
    else:
        return rfftn(ds_image_or_volume)


def _downsample_complex_signal_to_shape(
    image_or_volume: Complex[Array, "_ _"] | Complex[Array, "_ _ _"],
    downsampled_shape: tuple[int, int] | tuple[int, int, int],
    outputs_real_space: bool = True,
) -> Complex[Array, "_ _"] | Complex[Array, "_ _ _"]:
    fourier_array = jnp.fft.fftshift(fftn(image_or_volume))

    # Crop to the desired shape
    cropped_fourier_array = crop_to_shape(fourier_array, downsampled_shape)

    if outputs_real_space:
        return ifftn(jnp.fft.ifftshift(cropped_fourier_array))
    else:
        return jnp.fft.ifftshift(cropped_fourier_array)
